# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2016 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import traceback

from PathScripts.PathUtils import loopdetect
from PathScripts.PathUtils import horizontalEdgeLoop
from PathScripts.PathUtils import tangentEdgeLoop
from PathScripts.PathUtils import horizontalFaceLoop
from PathScripts.PathUtils import addToJob
from PathScripts.PathUtils import findParentJob

from PySide.QtCore import QT_TRANSLATE_NOOP

if FreeCAD.GuiUp:
    import FreeCADGui
    from PySide import QtCore
    from PySide import QtGui

# translate = FreeCAD.Qt.translate

__title__ = "FreeCAD Path Commands"
__author__ = "sliptonic"
__url__ = "https://www.freecad.org"


class _CommandSelectLoop:
    "the Path command to complete loop selection definition"

    def __init__(self):
        self.obj = None
        self.sub = []
        self.active = False

    def GetResources(self):
        return {
            "Pixmap": "CAM_SelectLoop",
            "MenuText": QT_TRANSLATE_NOOP("CAM_SelectLoop", "Finish Selecting Loop"),
            "Accel": "P, L",
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_SelectLoop",
                "Completes the selection of edges that form a loop\n Select one edge to search loop edges in horizontal plane\n Select two edges to search loop edges in wires of the shape\n Select one or more vertical faces to search loop faces which form the walls",
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if bool(FreeCADGui.Selection.getSelection()) is False:
            return False
        try:
            sel = FreeCADGui.Selection.getSelectionEx()[0]
            if sel.Object == self.obj and sel.SubElementNames == self.sub:
                return self.active
            self.obj = sel.Object
            self.sub = sel.SubElementNames
            if sel.SubObjects:
                # self.active = self.formsPartOfALoop(sel.Object, sel.SubObjects[0], sel.SubElementNames)
                self.active = True
            else:
                self.active = False
            return self.active
        except Exception as exc:
            Path.Log.error(exc)
            traceback.print_exc(exc)
            return False

    def Activated(self):
        from PathScripts.PathUtils import horizontalEdgeLoop, horizontalFaceLoop

        if not FreeCADGui.Selection.getSelectionEx():
            return

        sel = FreeCADGui.Selection.getSelectionEx()[0]
        if not sel.SubObjects:
            return

        obj = sel.Object
        sub = sel.SubObjects
        names = sel.SubElementNames
        loop = None

        # Face selection
        if "Face" in names[0]:
            loop = horizontalFaceLoop(obj, sub[0], names)
            if loop:
                FreeCADGui.Selection.clearSelection()
                FreeCADGui.Selection.addSelection(obj, loop)
                return

        elif "Edge" in names[0]:
            if len(sub) == 1:
                # One edge selected
                loop = horizontalEdgeLoop(obj, sub[0], verbose=True)

            if len(sub) >= 2:
                # Several edges selected
                loop = loopdetect(obj, sub[0], sub[1])

            if not loop:
                # Try to find tangent non planar loop
                loop = tangentEdgeLoop(obj, sub[0])

        if isinstance(loop, list) and len(loop) > 0 and isinstance(loop[0], Part.Edge):
            # Select edges from list
            objEdges = obj.Shape.Edges
            FreeCADGui.Selection.clearSelection()
            for el in loop:
                for eo in objEdges:
                    if eo.hashCode() == el.hashCode():
                        FreeCADGui.Selection.addSelection(obj, f"Edge{objEdges.index(eo) + 1}")
            return

        # Final fallback
        if FreeCAD.GuiUp:
            QtGui.QMessageBox.information(
                None,
                QT_TRANSLATE_NOOP("CAM_SelectLoop", "Feature Completion"),
                QT_TRANSLATE_NOOP("CAM_SelectLoop", "Closed loop detection failed."),
            )

    def formsPartOfALoop(self, obj, sub, names):
        try:
            if names[0][0:4] != "Edge":
                if names[0][0:4] == "Face" and horizontalFaceLoop(obj, sub, names):
                    return True
                return False
            if len(names) == 1 and horizontalEdgeLoop(obj, sub):
                return True
            if len(names) == 1 or names[1][0:4] != "Edge":
                return False
            return True
        except Exception:
            return False


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_SelectLoop", _CommandSelectLoop())


class _ToggleOperation:
    "command definition to toggle Operation Active state"

    def GetResources(self):
        return {
            "Pixmap": "CAM_OpActive",
            "MenuText": QT_TRANSLATE_NOOP("CAM_OpActiveToggle", "Toggle Operation"),
            "Accel": "P, X",
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_OpActiveToggle", "Toggles the active state of the operation"
            ),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        selection = FreeCADGui.Selection.getSelection()
        if not selection:
            return False

        for sel in selection:
            baseOp = Path.Dressup.Utils.baseOp(sel)
            if not hasattr(baseOp, "Active"):
                return False

        return True

    def Activated(self):
        selection = FreeCADGui.Selection.getSelection()
        for sel in selection:
            baseOp = Path.Dressup.Utils.baseOp(sel)
            baseOp.Active = not baseOp.Active

        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_OpActiveToggle", _ToggleOperation())


class _CopyOperation:
    "the Path Copy Operation command definition"

    def GetResources(self):
        return {
            "Pixmap": "CAM_OpCopy",
            "MenuText": QT_TRANSLATE_NOOP("CAM_OperationCopy", "Copy Operation"),
            "ToolTip": QT_TRANSLATE_NOOP("CAM_OperationCopy", "Copies the operation in the job"),
            "CmdType": "ForEdit",
        }

    def IsActive(self):
        if bool(FreeCADGui.Selection.getSelection()) is False:
            return False
        try:
            for sel in FreeCADGui.Selection.getSelectionEx():
                if not isinstance(sel.Object.Proxy, Path.Op.Base.ObjectOp):
                    return False
            return True
        except (IndexError, AttributeError):
            return False

    def Activated(self):
        for sel in FreeCADGui.Selection.getSelectionEx():
            jobname = findParentJob(sel.Object).Name
            addToJob(FreeCAD.ActiveDocument.copyObject(sel.Object, False), jobname)


if FreeCAD.GuiUp:
    FreeCADGui.addCommand("CAM_OperationCopy", _CopyOperation())


# \c findShape() is referenced from Gui/Command.cpp and used by Path.Area commands.
# Do not remove!
def findShape(shape, subname=None, subtype=None):
    """To find a higher order shape containing the subshape with subname.
    E.g. to find the wire containing 'Edge1' in shape,
        findShape(shape,'Edge1','Wires')
    """
    if not subname:
        return shape
    ret = shape.getElement(subname)
    if not subtype or not ret or ret.isNull():
        return ret
    if subname.startswith("Face"):
        tp = "Faces"
    elif subname.startswith("Edge"):
        tp = "Edges"
    elif subname.startswith("Vertex"):
        tp = "Vertex"
    else:
        return ret
    for obj in getattr(shape, subtype):
        for sobj in getattr(obj, tp):
            if sobj.isEqual(ret):
                return obj
    return ret



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2024 The FreeCAD Project Association                    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides general utility functions used throughout the workbench.

This module contains auxiliary functions which can be used
in other modules of the workbench, and which don't require
the graphical user interface (GUI).
"""
## @package utils
# \ingroup draftutils
# \brief Provides general utility functions used throughout the workbench.

## \addtogroup draftutils
# @{
import os
import PySide.QtCore as QtCore

import FreeCAD as App
from draftutils import params
from draftutils.messages import _wrn, _err, _log
from draftutils.translate import translate
from builtins import open

# TODO: move the functions that require the graphical interface
# This module should not import any graphical commands; those should be
# in gui_utils
if App.GuiUp:
    import FreeCADGui as Gui
    import Draft_rc

    # The module is used to prevent complaints from code checkers (flake8)
    True if Draft_rc else False


ARROW_TYPES = ["Dot", "Circle", "Arrow", "Tick", "Tick-2", "None"]
DISPLAY_MODES = ["Flat Lines", "Shaded", "Wireframe", "Points"]
DRAW_STYLES = ["Solid", "Dashed", "Dotted", "Dashdot"]
arrowtypes = ARROW_TYPES


def get_default_annotation_style():
    arrow_start_type_index = params.get_param("dimsymbolstart")
    arrow_end_type_index = params.get_param("dimsymbolend")
    return {
        "ArrowSizeStart": ("float", params.get_param("arrowsizestart")),
        "ArrowSizeEnd": ("float", params.get_param("arrowsizeend")),
        "ArrowTypeStart": ("index", arrow_start_type_index, ARROW_TYPES[arrow_start_type_index]),
        "ArrowTypeEnd": ("index", arrow_end_type_index, ARROW_TYPES[arrow_end_type_index]),
        "Decimals": ("int", params.get_param("dimPrecision")),
        "DimOvershoot": ("float", params.get_param("dimovershoot")),
        "ExtLines": ("float", params.get_param("extlines")),
        "ExtOvershoot": ("float", params.get_param("extovershoot")),
        "FontName": ("font", params.get_param("textfont")),
        "FontSize": ("float", params.get_param("textheight")),
        "LineColor": ("color", params.get_param("DefaultAnnoLineColor") | 0x000000FF),
        "LineSpacing": ("float", params.get_param("LineSpacing")),
        "LineWidth": ("int", params.get_param("DefaultAnnoLineWidth")),
        "ScaleMultiplier": ("float", params.get_param("DefaultAnnoScaleMultiplier")),
        "ShowLine": ("bool", params.get_param("DimShowLine")),
        "ShowUnit": ("bool", params.get_param("showUnit")),
        "TextColor": ("color", params.get_param("DefaultTextColor") | 0x000000FF),
        "TextSpacing": ("float", params.get_param("dimspacing")),
        "UnitOverride": ("str", params.get_param("overrideUnit")),
    }


def repair_annotation_style(style):
    """Repair a V0.19, V0.20 or < V1.1 style.

    V0.19 and V0.20:
    Some properties were missing or misspelled.
    Some float values were wrongly stored as strings.

    V1.0 -> V1.1:
    ArrowSize has been replaced by ArrowSizeStart and ArrowSizeEnd.
    ArrowType has been replaced by ArrowTypeStart and ArrowTypeEnd.
    """
    for key in ("ArrowSize", "ArrowType"):
        if (
            style.get(key) is not None
            and style.get(key + "Start") is None
            and style.get(key + "End") is None
        ):
            style[key + "Start"] = style[key]
            style[key + "End"] = style[key]
    default = get_default_annotation_style()
    new = {}
    for key, val in default.items():
        if style.get(key) is None:
            new[key] = val[1]
        elif type(style[key]) == type(val[1]):
            new[key] = style[key]
        elif isinstance(style[key], str):
            new[key] = float(style[key].replace(",", "."))
        else:
            new[key] = val[1]
    return new


def get_default_shape_style():
    # Uses the same format as get_default_annotation_style().
    display_mode_index = params.get_param("DefaultDisplayMode")
    draw_style_index = params.get_param("DefaultDrawStyle")
    return {
        "DisplayMode": ("index", display_mode_index, DISPLAY_MODES[display_mode_index]),
        "DrawStyle": ("index", draw_style_index, DRAW_STYLES[draw_style_index]),
        "LineColor": ("color", params.get_param_view("DefaultShapeLineColor") | 0x000000FF),
        "LineWidth": ("int", params.get_param_view("DefaultShapeLineWidth")),
        "PointColor": ("color", params.get_param_view("DefaultShapeVertexColor") | 0x000000FF),
        "PointSize": ("int", params.get_param_view("DefaultShapePointSize")),
        "ShapeAppearance": ("material", (get_view_material(),)),
    }


def get_view_material():
    """Return a ShapeAppearance material with properties based on the preferences."""
    material = App.Material()
    material.AmbientColor = params.get_param_view("DefaultAmbientColor") | 0x000000FF
    material.DiffuseColor = params.get_param_view("DefaultShapeColor") | 0x000000FF
    material.EmissiveColor = params.get_param_view("DefaultEmissiveColor") | 0x000000FF
    material.Shininess = params.get_param_view("DefaultShapeShininess") / 100
    material.SpecularColor = params.get_param_view("DefaultSpecularColor") | 0x000000FF
    material.Transparency = params.get_param_view("DefaultShapeTransparency") / 100
    return material


def string_encode_coin(ustr):
    """Encode a unicode object to be used as a string in coin.

    Parameters
    ----------
    ustr : str
        A string to be encoded

    Returns
    -------
    str
        Encoded string. If the coin version is >= 4
        it will encode the string to `'utf-8'`, otherwise
        it will encode it to `'latin-1'`.
    """
    try:
        from pivy import coin

        coin4 = coin.COIN_MAJOR_VERSION >= 4
    except (ImportError, AttributeError):
        coin4 = False
    if coin4:
        return ustr.encode("utf-8")
    else:
        return ustr.encode("latin1")


stringencodecoin = string_encode_coin


def type_check(args_and_types, name="?"):
    """Check that the arguments are instances of certain types.

    Parameters
    ----------
    args_and_types : list
        A list of tuples. The first element of a tuple is tested as being
        an instance of the second element.
        ::
            args_and_types = [(a, Type), (b, Type2), ...]

        Then
        ::
            isinstance(a, Type)
            isinstance(b, Type2)

        A `Type` can also be a tuple of many types, in which case
        the check is done for any of them.
        ::
            args_and_types = [(a, (Type3, int, float)), ...]

            isinstance(a, (Type3, int, float))

    name : str, optional
        Defaults to `'?'`. The name of the check.

    Raises
    ------
    TypeError
        If the first element in the tuple is not an instance of the second
        element, it raises `Draft.name`.
    """
    for v, t in args_and_types:
        if not isinstance(v, t):
            w = "typecheck[{}]: '{}' is not {}".format(name, v, t)
            _err(w)
            raise TypeError("Draft." + str(name))


typecheck = type_check


def precision():
    """Return the precision value from the parameter database.

    It is the number of decimal places that a float will have.
    Example
    ::
        precision=6, 0.123456
        precision=5, 0.12345
        precision=4, 0.1234

    Due to floating point operations there may be rounding errors.
    Therefore, this precision number is used to round up values
    so that all operations are consistent.
    By default the precision is 6 decimal places.

    Returns
    -------
    int
        params.get_param("precision")
    """
    return params.get_param("precision")


def svg_precision():
    """Return the precision value for SVG import from the parameter database.

    It is the number of decimal places that a float will have.
    Example
    ::
        precision=5, 0.12345
        precision=4, 0.1234
        precision=3, 0.123

    Due to floating point operations there may be rounding errors.
    Therefore, this precision number is used to round up values
    so that all operations are consistent.
    By default the precision is 3 decimal places.

    Returns
    -------
    int
        params.get_param("svgPrecision")
    """
    return params.get_param("svgPrecision")


def tolerance():
    """Return a tolerance based on the precision() value

    Returns
    -------
    float
        10 ** -precision()
    """
    return 10 ** -precision()


def is_deleted(obj):
    """Return `True` if obj is deleted."""
    try:
        return not obj.isAttachedToDocument()
    except:
        return True


def get_real_name(name):
    """Strip the trailing numbers from a string to get only the letters.

    Parameters
    ----------
    name : str
        A string that may have a number at the end, `Line001`.

    Returns
    -------
    str
        A string without the numbers at the end, `Line`.
        The returned string cannot be empty; it will have
        at least one letter.
    """
    for i in range(1, len(name) + 1):
        if name[-i] not in "1234567890":
            return name[: len(name) - (i - 1)]
    return name


getRealName = get_real_name


def get_type(obj):
    """Return a string indicating the type of the given object.

    Parameters
    ----------
    obj : App::DocumentObject
        Any type of scripted object created with Draft,
        or any other workbench.

    Returns
    -------
    str
        If `obj` has a `Proxy`, it will return the value of `obj.Proxy.Type`.

        * If `obj` is a `Part.Shape`, returns `'Shape'`

        * If `obj` has a `TypeId`, returns `obj.TypeId`

        In other cases, it will return `'Unknown'`,
        or `None` if `obj` is `None`.
    """
    import Part

    if not obj:
        return None
    if isinstance(obj, Part.Shape):
        return "Shape"
    if hasattr(obj, "Class") and "Ifc" in str(obj.Class):
        return obj.Class
    if hasattr(obj, "Proxy") and hasattr(obj.Proxy, "Type"):
        return obj.Proxy.Type
    if hasattr(obj, "TypeId"):
        return obj.TypeId
    return "Unknown"


getType = get_type


def get_objects_of_type(objects, typ):
    """Return only the objects that match the type in the list of objects.

    Parameters
    ----------
    objects : list of App::DocumentObject
        A list of objects which will be tested.

    typ : str
        A string that indicates a type. This should be one of the types
        that can be returned by `get_type`.

    Returns
    -------
    list of objects
        Only the objects that match `typ` will be added to the output list.
    """
    objs = []
    for o in objects:
        if getType(o) == typ:
            objs.append(o)
    return objs


getObjectsOfType = get_objects_of_type


def is_clone(obj, objtype=None, recursive=False):
    """Return True if the given object is a clone of a certain type.

    A clone is of type `'Clone'`, and has a reference
    to the original object inside its `Objects` attribute,
    which is an `'App::PropertyLinkListGlobal'`.

    The `Objects` attribute can point to another `'Clone'` object.
    If `recursive` is `True`, the function will be called recursively
    to further test this clone, until the type of the original object
    can be compared to `objtype`.

    Parameters
    ----------
    obj : App::DocumentObject
        The clone object that will be tested for a certain type.

    objtype : str or list of str
        A type string such as one obtained from `get_type`.
        Or a list of such types.

    recursive : bool, optional
        It defaults to `False`.
        If it is `True`, this same function will be called recursively
        with `obj.Object[0]` as input.

        This option only works if `obj.Object[0]` is of type `'Clone'`,
        that is, if `obj` is a clone of a clone.

    Returns
    -------
    bool
        Returns `True` if `obj` is of type `'Clone'`,
        and `obj.Object[0]` is of type `objtype`.

        If `objtype` is a list, then `obj.Objects[0]`
        will be tested against each of the elements in the list,
        and it will return `True` if at least one element matches the type.

        If `obj` isn't of type `'Clone'` but has the `CloneOf` attribute,
        it will also return `True`.

        It returns `False` otherwise, for example,
        if `obj` is not even a clone.
    """
    if isinstance(objtype, list):
        return any([is_clone(obj, t, recursive) for t in objtype])
    if getType(obj) == "Clone":
        if len(obj.Objects) == 1:
            if objtype:
                if getType(obj.Objects[0]) == objtype:
                    return True
            elif recursive and (getType(obj.Objects[0]) == "Clone"):
                return is_clone(obj.Objects[0], objtype, recursive)
    elif hasattr(obj, "CloneOf"):
        if obj.CloneOf:
            if objtype:
                if getType(obj.CloneOf) == objtype:
                    return True
            else:
                return True
    return False


isClone = is_clone


def get_clone_base(obj, strict=False, recursive=True):
    """Return the object cloned by this object, if any.

    Parameters
    ----------
    obj: App::DocumentObject
        Any type of object.

    strict: bool, optional
        It defaults to `False`.
        If it is `True`, and this object is not a clone,
        this function will return `False`.

    recursive: bool, optional
        It defaults to `True`
        If it is `True`, it call recursively to itself to
        get base object and if it is `False` then it just
        return base object, not call recursively to find
        base object.

    Returns
    -------
    App::DocumentObject
        It `obj` is a `Draft Clone`, it will return the first object
        that is in its `Objects` property.

        If `obj` has a `CloneOf` property, it will search iteratively
        inside the object pointed to by this property.

    obj
        If `obj` is not a `Draft Clone`, nor it has a `CloneOf` property,
        it will return the same `obj`, as long as `strict` is `False`.

    False
        It will return `False` if `obj` is not a clone,
        and `strict` is `True`.
    """
    if hasattr(obj, "CloneOf") and obj.CloneOf:
        if recursive:
            return get_clone_base(obj.CloneOf)
        return obj.CloneOf
    if get_type(obj) == "Clone" and obj.Objects:
        if recursive:
            return get_clone_base(obj.Objects[0])
        return obj.Objects[0]
    if strict:
        return False
    return obj


getCloneBase = get_clone_base


def shapify(obj, delete=True):
    """Transform a parametric object into a static, non-parametric shape.

    Parameters
    ----------
    obj : App::DocumentObject
        Any type of scripted object.

        This object will be removed, and a non-parametric object
        with the same topological shape (`Part::TopoShape`)
        will be created.

    delete: bool, optional
        It defaults to `False`.
        If it is `True`, the original object is deleted.

    Returns
    -------
    Part::Feature
        The new object that takes `obj.Shape` as its own.

        Depending on the contents of the Shape, the resulting object
        will be named `'Face'`, `'Solid'`, `'Compound'`,
        `'Shell'`, `'Wire'`, `'Line'`, `'Circle'`,
        or the name returned by `get_real_name(obj.Name)`.

        If there is a problem with `obj.Shape`, it will return `None`,
        and the original object will not be modified.
    """
    try:
        shape = obj.Shape
    except Exception:
        return None

    if len(shape.Faces) == 1:
        name = "Face"
    elif len(shape.Solids) == 1:
        name = "Solid"
    elif len(shape.Solids) > 1:
        name = "Compound"
    elif len(shape.Faces) > 1:
        name = "Shell"
    elif len(shape.Wires) == 1:
        name = "Wire"
    elif len(shape.Edges) == 1:
        import DraftGeomUtils

        if DraftGeomUtils.geomType(shape.Edges[0]) == "Line":
            name = "Line"
        else:
            name = "Circle"
    else:
        name = getRealName(obj.Name)

    if delete:
        App.ActiveDocument.removeObject(obj.Name)
    newobj = App.ActiveDocument.addObject("Part::Feature", name)
    newobj.Shape = shape

    return newobj


def print_shape(shape):
    """Print detailed information of a topological shape.

    Parameters
    ----------
    shape : Part::TopoShape
        Any topological shape in an object, usually obtained from `obj.Shape`.
    """
    _msg(translate("draft", "Solids:") + " {}".format(len(shape.Solids)))
    _msg(translate("draft", "Faces:") + " {}".format(len(shape.Faces)))
    _msg(translate("draft", "Wires:") + " {}".format(len(shape.Wires)))
    _msg(translate("draft", "Edges:") + " {}".format(len(shape.Edges)))
    _msg(translate("draft", "Vertices:") + " {}".format(len(shape.Vertexes)))

    if shape.Faces:
        for f in range(len(shape.Faces)):
            _msg(translate("draft", "Face") + " {}:".format(f))
            for v in shape.Faces[f].Vertexes:
                _msg("    {}".format(v.Point))
    elif shape.Wires:
        for w in range(len(shape.Wires)):
            _msg(translate("draft", "Wire") + " {}:".format(w))
            for v in shape.Wires[w].Vertexes:
                _msg("    {}".format(v.Point))
    else:
        for v in shape.Vertexes:
            _msg("    {}".format(v.Point))


printShape = print_shape


def compare_objects(obj1, obj2):
    """Print the differences between 2 objects.

    The two objects are compared through their `TypeId` attribute,
    or by using the `get_type` function.

    If they are the same type their properties are compared
    looking for differences.

    Neither `Shape` nor `Label` attributes are compared.

    Parameters
    ----------
    obj1 : App::DocumentObject
        Any type of scripted object.
    obj2 : App::DocumentObject
        Any type of scripted object.
    """
    if obj1.TypeId != obj2.TypeId:
        _msg(
            "'{0}' ({1}), '{2}' ({3}): ".format(obj1.Name, obj1.TypeId, obj2.Name, obj2.TypeId)
            + translate("draft", "different types")
            + " (TypeId)"
        )
    elif getType(obj1) != getType(obj2):
        _msg(
            "'{0}' ({1}), '{2}' ({3}): ".format(
                obj1.Name, get_type(obj1), obj2.Name, get_type(obj2)
            )
            + translate("draft", "different types")
            + " (Proxy.Type)"
        )
    else:
        for p in obj1.PropertiesList:
            if p in obj2.PropertiesList:
                if p in ("Shape", "Label"):
                    pass
                elif p == "Placement":
                    delta = obj1.Placement.Base.sub(obj2.Placement.Base)
                    text = translate(
                        "draft",
                        "Objects have different placements. "
                        "Distance between the two base points:",
                    )
                    _msg(text + " " + str(delta.Length))
                else:
                    if getattr(obj1, p) != getattr(obj2, p):
                        _msg("'{}' ".format(p) + translate("draft", "has a different value"))
            else:
                _msg("{} ".format(p) + translate("draft", "doesn't exist in one of the objects"))


compareObjects = compare_objects


def load_svg_patterns():
    """Load the default Draft SVG patterns and user defined patterns.

    The SVG patterns are added as a dictionary to the `App.svgpatterns`
    attribute.
    """
    import importSVG

    App.svgpatterns = {}

    # Get default patterns in the resource file
    patfiles = QtCore.QDir(":/patterns").entryList()
    for fn in patfiles:
        file = ":/patterns/" + str(fn)
        f = QtCore.QFile(file)
        f.open(QtCore.QIODevice.ReadOnly)
        p = importSVG.getContents(str(f.readAll()), "pattern", True)
        if p:
            for k in p:
                p[k] = [p[k], file]
            App.svgpatterns.update(p)

    # Get patterns in a user defined file
    altpat = params.get_param("patternFile")
    if os.path.isdir(altpat):
        for f in os.listdir(altpat):
            if f[-4:].upper() == ".SVG":
                file = os.path.join(altpat, f)
                p = importSVG.getContents(file, "pattern")
                if p:
                    for k in p:
                        p[k] = [p[k], file]
                    App.svgpatterns.update(p)

    # Get TechDraw patterns
    altpat = os.path.join(App.getResourceDir(), "Mod", "TechDraw", "Patterns")
    if os.path.isdir(altpat):
        for f in os.listdir(altpat):
            if f[-4:].upper() == ".SVG":
                file = os.path.join(altpat, f)
                p = importSVG.getContents(file, "pattern")
                if p:
                    for k in p:
                        p[k] = [p[k], file]
                else:
                    # some TD pattern files have no <pattern> definition but can still be used by Draft
                    p = {f[:-4]: ["<pattern></pattern>", file]}
                    App.svgpatterns.update(p)


loadSvgPatterns = load_svg_patterns


def svg_patterns():
    """Return a dictionary with installed SVG patterns.

    Returns
    -------
    dict
        Returns `App.svgpatterns` if it exists.
        Otherwise it calls `load_svg_patterns` to create it
        before returning it.
    """
    if hasattr(App, "svgpatterns"):
        return App.svgpatterns
    else:
        loadSvgPatterns()
        if hasattr(App, "svgpatterns"):
            return App.svgpatterns
    return {}


svgpatterns = svg_patterns


def get_rgb(color, testbw=True):
    """Return an RRGGBB value #000000 from a FreeCAD color.

    Parameters
    ----------
    color : list or tuple with RGB values
        The values must be in the 0.0-1.0 range.
    testwb : bool (default = True)
        Pure white will be converted into pure black.
    """
    r = str(hex(int(color[0] * 255)))[2:].zfill(2)
    g = str(hex(int(color[1] * 255)))[2:].zfill(2)
    b = str(hex(int(color[2] * 255)))[2:].zfill(2)
    col = "#" + r + g + b
    if testbw:
        if col == "#ffffff":
            # print(params.get_param("SvgLinesBlack"))
            if params.get_param("SvgLinesBlack"):
                col = "#000000"
    return col


getrgb = get_rgb


def argb_to_rgba(color):
    """Change byte order of a 4 byte color int from ARGB (Qt) to RGBA (FreeCAD).

    Alpha in both integers should always be 255.

    Alpha in color properties is not used in the 3D view, but is shown in the
    color swatches in the Property editor. It therefore better to ensure alpha
    is 255 (version 1.1 dev cycle).

    Usage:

        qt_int = self.form.ShapeColor.property("color").rgba() # Note: returns ARGB int
        qt_int = self.form.ShapeColor.property("color").rgb()  # Note: returns ARGB int
        fc_int = argb_to_rgba(qt_int)

        FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")\
            .SetUnsigned("DefaultShapeColor", fc_int)

        obj.ViewObject.ShapeColor = fc_int | 0x000000FF

    Related:

        getRgbF() returns an RGBA tuple. 4 floats in the range 0.0 - 1.0. Alpha is always 1.0.
    """
    return ((color & 0xFFFFFF) << 8) + ((color & 0xFF000000) >> 24)


def rgba_to_argb(color):
    """Change byte order of a 4 byte color int from RGBA (FreeCAD) to ARGB (Qt)."""
    return ((color & 0xFFFFFF00) >> 8) + ((color & 0xFF) << 24)


def get_rgba_tuple(color, typ=1.0):
    """Return an RGBA tuple.

    Parameters
    ----------
    color: int
        RGBA integer.
    typ: any float (default = 1.0) or int (use 255)
        If float the values in the returned tuple are in the 0.0-1.0 range.
        Else the values are in the 0-255 range.
    """
    color = ((color >> 24) & 0xFF, (color >> 16) & 0xFF, (color >> 8) & 0xFF, color & 0xFF)
    if type(typ) == float:
        return tuple([x / 255.0 for x in color])
    else:
        return color


def _modifiers_process_subselection(sels, copy):
    data_list = []
    sel_info = []
    for sel in sels:
        for sub in sel.SubElementNames if sel.SubElementNames else [""]:
            if not ("Vertex" in sub or "Edge" in sub):
                continue
            if copy and "Vertex" in sub:
                continue
            obj = sel.Object.getSubObject(sub, 1)
            pla = sel.Object.getSubObject(sub, 3)
            if "Vertex" in sub:
                vert_idx = int(sub.rpartition("Vertex")[2]) - 1
                edge_idx = -1
            else:
                vert_idx = -1
                edge_idx = int(sub.rpartition("Edge")[2]) - 1
            data_list.append((obj, vert_idx, edge_idx, pla))
            sel_info.append(("", sel.Object.Name, sub))
    return data_list, sel_info


def _modifiers_process_selection(sels, copy, scale=False, add_movable_children=False):
    # Only when creating ghosts and if copy is False, should add_movable_children be True.
    objects = []
    places = []
    sel_info = []
    for sel in sels:
        for sub in sel.SubElementNames if sel.SubElementNames else [""]:
            obj = sel.Object.getSubObject(sub, 1)
            # Get the global placement of the parent:
            if obj == sel.Object:
                pla = App.Placement()
            else:
                pla = sel.Object.getSubObject(sub.rpartition(obj.Name)[0], 3)
            objs = _modifiers_get_group_contents(obj)
            if add_movable_children:
                children = []
                for obj in objs:
                    children.extend(_modifiers_get_movable_children(obj))
                objs.extend(children)
            objs = _modifiers_filter_objects(objs, copy, scale)
            objects.extend(objs)
            places.extend(len(objs) * [pla])
            if "." in sub:
                sub = sub.rpartition(".")[0] + "."
            elif "Face" in sub or "Edge" in sub or "Vertex" in sub:
                sub = ""
            sel_info.append(("", sel.Object.Name, sub))
    return objects, places, sel_info


def _modifiers_get_group_contents(obj):
    from draftutils import groups

    return groups.get_group_contents(obj, addgroups=True, spaces=True, noarchchild=True)


def _modifiers_get_movable_children(obj):
    result = []
    if hasattr(obj, "Proxy") and hasattr(obj.Proxy, "getMovableChildren"):
        children = obj.Proxy.getMovableChildren(obj)
        result.extend(children)
        for child in children:
            result.extend(_modifiers_get_movable_children(child))
    return result


def _modifiers_filter_objects(objs, copy, scale=False):

    def is_scalable(obj):
        if hasattr(obj, "Placement") and hasattr(obj, "Shape"):
            return True
        if obj.isDerivedFrom("App::DocumentObjectGroup"):
            return True
        if obj.isDerivedFrom("App::Annotation"):
            return True
        if obj.isDerivedFrom("Image::ImagePlane"):
            return True
        return False

    result = []
    for obj in objs:
        if not copy and hasattr(obj, "MoveBase") and obj.MoveBase and obj.Base:
            parents = []
            for parent in obj.Base.InList:
                if parent.isDerivedFrom("Part::Feature"):
                    parents.append(parent.Name)
            if len(parents) > 1:
                message = translate(
                    "draft",
                    "%s shares a base with %d other objects. Please check if you want to modify this.",
                ) % (obj.Name, len(parents) - 1)
                _err(message)
            if not scale or utils.get_type(obj.Base) == "Wire":
                result.append(obj.Base)
        elif (
            not copy and hasattr(obj, "Placement") and "ReadOnly" in obj.getEditorMode("Placement")
        ):
            _err(
                translate("draft", "%s cannot be modified because its placement is readonly")
                % obj.Name
            )
        elif not scale or is_scalable(obj):
            result.append(obj)
    return result


def is_closed_edge(edge_index, object):
    return edge_index + 1 >= len(object.Points)


def utf8_decode(text):
    r"""Decode the input string and return a unicode string.

    Python 2:
    ::
        str -> unicode
        unicode -> unicode

    Python 3:
    ::
        str -> str
        bytes -> str

    It runs
    ::
        try:
            return text.decode("utf-8")
        except AttributeError:
            return text

    Parameters
    ----------
    text : str, unicode or bytes
        A str, unicode, or bytes object that may have unicode characters
        like accented characters.

        In Python 2, a `bytes` object can include accented characters,
        but in Python 3 it must only contain ASCII literal characters.

    Returns
    -------
    unicode or str
        In Python 2 it will try decoding the `bytes` string
        and return a `'utf-8'` decoded string.

        >>> "Aá".decode("utf-8")
        >>> b"Aá".decode("utf-8")
        u'A\xe1'

        In Python 2 the unicode string is prefixed with `u`,
        and unicode characters are replaced by their two-digit hexadecimal
        representation, or four digit unicode escape.

        >>> "AáBẃCñ".decode("utf-8")
        u'A\xe1B\u1e83C\xf1'

        In Python 2 it will always return a `unicode` object.

        In Python 3 a regular string is already unicode encoded,
        so strings have no `decode` method. In this case, `text`
        will be returned as is.

        In Python 3, if `text` is a `bytes` object, then it will be converted
        to `str`; in this case, the `bytes` object cannot have accents,
        it must only contain ASCII literal characters.

        >>> b"ABC".decode("utf-8")
        'ABC'

        In Python 3 it will always return a `str` object, with no prefix.
    """
    try:
        return text.decode("utf-8")
    except AttributeError:
        return text


def print_header(name, description, debug=True):
    """Print a line to the console when something is called, and log it.

    Parameters
    ----------
    name: str
        The name of the function or class that is being called.
        This `name` will be logged in the log file, so if there are problems
        the log file can be investigated for clues.

    description: str
        Arbitrary text that will be printed to the console
        when the function or class is called.

    debug: bool, optional
        It defaults to `True`.
        If it is `False` the `description` will not be printed
        to the console.
        On the other hand the `name` will always be logged.
    """
    _log(name)
    if debug:
        _msg(16 * "-")
        _msg(description)


def find_doc(doc=None):
    """Return the active document or find a document by name.

    Parameters
    ----------
    doc: App::Document or str, optional
        The document that will be searched in the session.
        It defaults to `None`, in which case it tries to find
        the active document.
        If `doc` is a string, it will try to get the document by `Name`.

    Returns
    -------
    bool, App::Document
        A tuple containing the information on whether the search
        was successful. In this case, the boolean is `True`,
        and the second value is the document instance.

    False, None
        If there is no active document, or the string in `doc`
        doesn't correspond to an open document in the session.
    """
    FOUND = True

    if not doc:
        doc = App.activeDocument()
    if not doc:
        return not FOUND, None

    if isinstance(doc, str):
        try:
            doc = App.getDocument(doc)
        except NameError:
            _err(translate("draft", "Wrong input: unknown document {}").format(doc))
            return not FOUND, None

    return FOUND, doc


def find_object(obj, doc=None):
    """Find object in the document, inclusive by Label.

    Parameters
    ----------
    obj: App::DocumentObject or str
        The object to search in `doc`.
        Or if the `obj` is a string, it will search the object by `Label`.
        Since Labels are not guaranteed to be unique, it will get the first
        object with that label in the document.

    doc: App::Document or str, optional
        The document in which the object will be searched.
        It defaults to `None`, in which case it tries to search in the
        active document.
        If `doc` is a string, it will search the document by `Name`.

    Returns
    -------
    bool, App::DocumentObject
        A tuple containing the information on whether the search
        was successful. In this case, the boolean is `True`,
        and the second value is the object found.

    False, None
        If the object doesn't exist in the document.
    """
    FOUND = True

    found, doc = find_doc(doc)
    if not found:
        _err(translate("draft", "No active document. Aborting."))
        return not FOUND, None

    if isinstance(obj, str):
        try:
            obj = doc.getObjectsByLabel(obj)[0]
        except IndexError:
            return not FOUND, None

    if obj not in doc.Objects:
        return not FOUND, None

    return FOUND, obj


def use_instead(function, version=""):
    """Print a deprecation message and suggest another function.

    This function must be used inside the definition of a function
    that has been considered for deprecation, so we must provide
    an alternative.
    ::
        def old_function():
            use_instead('new_function', 1.0)

        def someFunction():
            use_instead('some_function')

    Parameters
    ----------
    function: str
        The name of the function to use instead of the current one.

    version: float or str, optional
        It defaults to the empty string `''`.
        The version where this command is to be deprecated, if it is known.
        If we don't know when this command will be deprecated
        then we should not give a version.
    """
    if version:
        _wrn(
            translate("draft", "This function will be deprecated in {}. Please use '{}'.").format(
                version, function
            )
        )
    else:
        _wrn(
            translate("draft", "This function will be deprecated. Please use '{}'.").format(
                function
            )
        )


def pyopen(
    file,
    mode="r",
    buffering=-1,
    encoding=None,
    errors=None,
    newline=None,
    closefd=True,
    opener=None,
):
    if encoding is None:
        encoding = "utf-8"
    return open(file, mode, buffering, encoding, errors, newline, closefd, opener)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Modules that define classes that handle task panels of the GuiCommands.

These classes load `.ui` files that will be used in the task panels
of the graphical commands defined in the `draftguitools` submodules.
These classes also define the callbacks and interactions of the widgets
in the `.ui` files.

The task panel interface should be, in the possible, not manually written
in Python. Rather a `.ui` file should be built using Qt Designer,
and should be placed in the `Resources/ui/` directory so that it is loaded
by the respective task panel class.

::

    class CommandTaskPanel:
        def __init__(self):
            ui_file = ":/ui/TaskPanel_OrthoArray.ui"
            self.form = Gui.PySideUic.loadUi(ui_file)

At the moment, most GuiCommands don't have their own `.ui` file,
nor task panel class. Instead their task panels are completely defined
by the `DraftGui` module.

This should be changed in the future because `DraftGui`
is a very large module, and creates many different widgets
and handles many callbacks, making it difficult to extend and maintain.
This module selectively chooses the widgets to show and to hide
depending on the command that is activated.

Individual task panel classes and `.ui` files are more maintainable
because changes can be done to a single tool without affecting the rest,
and the module size is kept small.
"""
## \defgroup drafttaskpanels drafttaskpanels
# \ingroup DRAFT
# \brief Modules with classes that handle task panels of the GuiCommands.



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Wall"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchWall
#  \ingroup ARCH
#  \brief The Wall object and tools
#
#  This module provides tools to build Wall objects.  Walls are simple objects,
#  usually vertical, typically obtained by giving a thickness to a base line,
#  then extruding it vertically.

"""This module provides tools to build Wall objects.  Walls are simple
objects, usually vertical, typically obtained by giving a thickness to a base
line, then extruding it vertically.

Examples
--------
TODO put examples here.

"""

import math

import FreeCAD
import ArchCommands
import ArchComponent
import ArchSketchObject
import Draft
import DraftVecUtils

from FreeCAD import Vector
from draftutils import params

if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    import draftguitools.gui_trackers as DraftTrackers
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


def mergeShapes(w1, w2):
    """Not currently implemented.

    Return a Shape built on two walls that share same properties and have a
    coincident endpoint.
    """

    if not areSameWallTypes([w1, w2]):
        return None
    if (not hasattr(w1.Base, "Shape")) or (not hasattr(w2.Base, "Shape")):
        return None
    if w1.Base.Shape.Faces or w2.Base.Shape.Faces:
        return None

    # TODO fix this
    return None

    eds = w1.Base.Shape.Edges + w2.Base.Shape.Edges
    import DraftGeomUtils

    w = DraftGeomUtils.findWires(eds)
    if len(w) == 1:
        # print("found common wire")
        normal, length, width, height = w1.Proxy.getDefaultValues(w1)
        print(w[0].Edges)
        sh = w1.Proxy.getBase(w1, w[0], normal, width, height)
        print(sh)
        return sh
    return None


def areSameWallTypes(walls):
    """Check if a list of walls have the same height, width and alignment.

    Parameters
    ----------
    walls: list of <ArchComponent.Component>

    Returns
    -------
    bool
        True if the walls have the same height, width and alignment, False if
        otherwise.
    """

    for att in ["Width", "Height", "Align"]:
        value = None
        for w in walls:
            if not hasattr(w, att):
                return False
            if not value:
                value = getattr(w, att)
            else:
                if type(value) == float:
                    if round(value, Draft.precision()) != round(getattr(w, att), Draft.precision()):
                        return False
                else:
                    if value != getattr(w, att):
                        return False
    return True


class _Wall(ArchComponent.Component):
    """The Wall object.

    Turns a <App::FeaturePython> into a wall object, then uses a
    <Part::Feature> to create the wall's shape.

    Walls are simple objects, usually vertical, typically obtained by giving a
    thickness to a base line, then extruding it vertically.

    Parameters
    ----------
    obj: <App::FeaturePython>
        The object to turn into a wall. Note that this is not the object that
        forms the basis for the new wall's shape. That is given later.
    """

    def __init__(self, obj):
        ArchComponent.Component.__init__(self, obj)
        self.Type = "Wall"
        self.setProperties(obj)
        obj.IfcType = "Wall"

    def setProperties(self, obj):
        """Give the wall its wall specific properties, such as its alignment.

        You can learn more about properties here:
        https://wiki.freecad.org/property

        parameters
        ----------
        obj: <part::featurepython>
            The object to turn into a wall.
        """

        lp = obj.PropertiesList
        if not "Length" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "Length",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The length of this wall. Read-only if this wall is not based on an unconstrained sketch with a single edge, or on a Draft Wire with a single edge. Refer to wiki for details how length is deduced.",
                ),
                locked=True,
            )
        if not "Width" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "Width",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The width of this wall. Not used if this wall is based on a face. Disabled and ignored if Base object (ArchSketch) provides the information.",
                ),
                locked=True,
            )

        # To be combined into Width when PropertyLengthList is available
        if not "OverrideWidth" in lp:
            obj.addProperty(
                "App::PropertyFloatList",
                "OverrideWidth",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This overrides Width attribute to set width of each segment of wall.  Disabled and ignored if Base object (ArchSketch) provides Widths information, with getWidths() method  (If a value is zero, the value of 'Width' will be followed).  [ENHANCEMENT by ArchSketch] GUI 'Edit Wall Segment Width' Tool is provided in external SketchArch Add-on to let users to set the values interactively.  'Toponaming-Tolerant' if ArchSketch is used in Base (and SketchArch Add-on is installed).  Warning : Not 'Toponaming-Tolerant' if just Sketch is used.",
                ),
                locked=True,
            )  # see DraftGeomUtils.offsetwire()
        if not "OverrideAlign" in lp:
            obj.addProperty(
                "App::PropertyStringList",
                "OverrideAlign",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This overrides Align attribute to set align of each segment of wall.  Disabled and ignored if Base object (ArchSketch) provides Aligns information, with getAligns() method  (If a value is not 'Left, Right, Center', the value of 'Align' will be followed).  [ENHANCEMENT by ArchSketch] GUI 'Edit Wall Segment Align' Tool is provided in external SketchArch Add-on to let users to set the values interactively.  'Toponaming-Tolerant' if ArchSketch is used in Base (and SketchArch Add-on is installed).  Warning : Not 'Toponaming-Tolerant' if just Sketch is used.",
                ),
                locked=True,
            )  # see DraftGeomUtils.offsetwire()
        if not "OverrideOffset" in lp:
            obj.addProperty(
                "App::PropertyFloatList",
                "OverrideOffset",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "This overrides Offset attribute to set offset of each segment of wall.  Disabled and ignored if Base object (ArchSketch) provides Offsets information, with getOffsets() method  (If a value is zero, the value of 'Offset' will be followed).  [ENHANCED by ArchSketch] GUI 'Edit Wall Segment Offset' Tool is provided in external Add-on ('SketchArch') to let users to select the edges interactively.  'Toponaming-Tolerant' if ArchSketch is used in Base (and SketchArch Add-on is installed).  Warning : Not 'Toponaming-Tolerant' if just Sketch is used. Property is ignored if Base ArchSketch provided the selected edges. ",
                ),
                locked=True,
            )  # see DraftGeomUtils.offsetwire()
        if not "Height" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The height of this wall. Keep 0 for automatic. Not used if this wall is based on a solid",
                ),
                locked=True,
            )
        if not "Area" in lp:
            obj.addProperty(
                "App::PropertyArea",
                "Area",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The area of this wall as a simple Height * Length calculation"
                ),
                locked=True,
            )
            obj.setEditorMode("Area", 1)
        if not "Align" in lp:
            obj.addProperty(
                "App::PropertyEnumeration",
                "Align",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The alignment of this wall on its base object, if applicable. Disabled and ignored if Base object (ArchSketch) provides the information.",
                ),
                locked=True,
            )
            obj.Align = ["Left", "Right", "Center"]
        if not "Normal" in lp:
            obj.addProperty(
                "App::PropertyVector",
                "Normal",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The normal extrusion direction of this object (keep (0,0,0) for automatic normal)",
                ),
                locked=True,
            )
        if not "Face" in lp:
            obj.addProperty(
                "App::PropertyInteger",
                "Face",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The face number of the base object used to build this wall"
                ),
                locked=True,
            )
        if not "Offset" in lp:
            obj.addProperty(
                "App::PropertyDistance",
                "Offset",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "The offset between this wall and its baseline (only for left and right alignments). Disabled and ignored if Base object (ArchSketch) provides the information.",
                ),
                locked=True,
            )

        # See getExtrusionData(), removeSplitters are no longer used
        # if not "Refine" in lp:
        #    obj.addProperty("App::PropertyEnumeration","Refine","Wall",QT_TRANSLATE_NOOP("App::Property","Select whether or not and the method to remove splitter of the Wall. Currently Draft removeSplitter and Part removeSplitter available but may not work on complex sketch."), locked=True)
        #    obj.Refine = ['No','DraftRemoveSplitter','PartRemoveSplitter']
        # TODO - To implement in Arch Component ?

        if not "MakeBlocks" in lp:
            obj.addProperty(
                "App::PropertyBool",
                "MakeBlocks",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "Enable this to make the wall generate blocks"),
                locked=True,
            )
        if not "BlockLength" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "BlockLength",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "The length of each block"),
                locked=True,
            )
        if not "BlockHeight" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "BlockHeight",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "The height of each block"),
                locked=True,
            )
        if not "OffsetFirst" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "OffsetFirst",
                "Blocks",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The horizontal offset of the first line of blocks"
                ),
                locked=True,
            )
        if not "OffsetSecond" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "OffsetSecond",
                "Blocks",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The horizontal offset of the second line of blocks"
                ),
                locked=True,
            )
        if not "Joint" in lp:
            obj.addProperty(
                "App::PropertyLength",
                "Joint",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "The size of the joints between each block"),
                locked=True,
            )
        if not "CountEntire" in lp:
            obj.addProperty(
                "App::PropertyInteger",
                "CountEntire",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "The number of entire blocks"),
                locked=True,
            )
            obj.setEditorMode("CountEntire", 1)
        if not "CountBroken" in lp:
            obj.addProperty(
                "App::PropertyInteger",
                "CountBroken",
                "Blocks",
                QT_TRANSLATE_NOOP("App::Property", "The number of broken blocks"),
                locked=True,
            )
            obj.setEditorMode("CountBroken", 1)
        if not "ArchSketchData" in lp:
            obj.addProperty(
                "App::PropertyBool",
                "ArchSketchData",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Use Base ArchSketch (if used) data (e.g. widths, aligns, offsets) instead of Wall's properties",
                ),
                locked=True,
            )
            obj.ArchSketchData = True
        if not "ArchSketchEdges" in lp:
            obj.addProperty(
                "App::PropertyStringList",
                "ArchSketchEdges",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Selected edges (or group of edges) of the base Sketch/ArchSketch, to use in creating the shape of this Arch Wall (instead of using all the Base Sketch/ArchSketch's edges by default).  Input are index numbers of edges or groups.  Disabled and ignored if Base object (ArchSketch) provides selected edges (as Wall Axis) information, with getWallBaseShapeEdgesInfo() method.  [ENHANCEMENT by ArchSketch] GUI 'Edit Wall Segment' Tool is provided in external SketchArch Add-on to let users to (de)select the edges interactively.  'Toponaming-Tolerant' if ArchSketch is used in Base (and SketchArch Add-on is installed).  Warning : Not 'Toponaming-Tolerant' if just Sketch is used.",
                ),
                locked=True,
            )
        if not hasattr(obj, "ArchSketchPropertySet"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "ArchSketchPropertySet",
                "Wall",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Select User Defined PropertySet to use in creating variant shape, layers of the Arch Wall with same ArchSketch ",
                ),
                locked=True,
            )
            obj.ArchSketchPropertySet = ["Default"]
        if not hasattr(self, "ArchSkPropSetPickedUuid"):  # 'obj.Proxy', 'self' not works ?
            self.ArchSkPropSetPickedUuid = ""
        if not hasattr(self, "ArchSkPropSetListPrev"):
            self.ArchSkPropSetListPrev = []
        self.connectEdges = []

    def dumps(self):
        dump = super().dumps()
        if not isinstance(dump, tuple):
            dump = (dump,)  # Python Tuple With One Item
        dump = dump + (self.ArchSkPropSetPickedUuid, self.ArchSkPropSetListPrev)
        return dump

    def loads(self, state):
        self.Type = "Wall"
        if state == None:
            return
        elif state[0] == "W":  # state[1] == 'a', behaviour before 2024.11.28
            return
        elif state[0] == "Wall":
            self.ArchSkPropSetPickedUuid = state[1]
            self.ArchSkPropSetListPrev = state[2]
        elif state[0] != "Wall":  # model before merging super.dumps/loads()
            self.ArchSkPropSetPickedUuid = state[0]
            self.ArchSkPropSetListPrev = state[1]

    def onDocumentRestored(self, obj):
        """Method run when the document is restored. Re-adds the Arch component, and Arch wall properties."""

        import DraftGeomUtils
        from draftutils.messages import _log

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

        # In V1.0 the handling of wall normals has changed. As a result existing
        # walls with their Normal set to [0, 0, 0], based on wires or faces with
        # a shape normal pointing towards -Z, would be extruded in that direction
        # instead of towards +Z as before. To avoid this their Normal property is
        # changed to [0, 0, 1].
        if (
            FreeCAD.ActiveDocument.getProgramVersion() < "0.22"
            and obj.Normal == Vector(0, 0, 0)
            and hasattr(obj.Base, "Shape")
            and not obj.Base.Shape.Solids
            and obj.Face == 0
            and not obj.Base.isDerivedFrom("Sketcher::SketchObject")
            and DraftGeomUtils.get_shape_normal(obj.Base.Shape) != Vector(0, 0, 1)
        ):
            obj.Normal = Vector(0, 0, 1)
            _log(
                "v1.0, "
                + obj.Name
                + ", "
                + "changed 'Normal' to [0, 0, 1] to preserve extrusion direction"
            )

        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj, "Width"):  # TODO need test?
                obj.setEditorMode("Width", ["ReadOnly"])
            if hasattr(obj, "Align"):
                obj.setEditorMode("Align", ["ReadOnly"])
            if hasattr(obj, "Offset"):
                obj.setEditorMode("Offset", ["ReadOnly"])
            if hasattr(obj, "OverrideWidth"):
                obj.setEditorMode("OverrideWidth", ["ReadOnly"])
            if hasattr(obj, "OverrideAlign"):
                obj.setEditorMode("OverrideAlign", ["ReadOnly"])
            if hasattr(obj, "OverrideOffset"):
                obj.setEditorMode("OverrideOffset", ["ReadOnly"])
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", ["ReadOnly"])
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", 0)
        else:
            if hasattr(obj, "Width"):
                obj.setEditorMode("Width", 0)
            if hasattr(obj, "Align"):
                obj.setEditorMode("Align", 0)
            if hasattr(obj, "Offset"):
                obj.setEditorMode("Offset", 0)
            if hasattr(obj, "OverrideWidth"):
                obj.setEditorMode("OverrideWidth", 0)
            if hasattr(obj, "OverrideAlign"):
                obj.setEditorMode("OverrideAlign", 0)
            if hasattr(obj, "OverrideOffset"):
                obj.setEditorMode("OverrideOffset", 0)
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", 0)
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", ["ReadOnly"])

    def execute(self, obj):
        """Method run when the object is recomputed.

        Extrude the wall from the Base shape if possible. Processe additions
        and subtractions. Assign the resulting shape as the shape of the wall.

        Add blocks if the MakeBlocks property is assigned. If the Base shape is
        a mesh, just copy the mesh.
        """

        if self.clone(obj):
            return

        # Wall can do without Base, validity to be tested in getExtrusionData()
        # Remarked out ensureBase() below
        # if not self.ensureBase(obj):
        #    return

        import Part
        import DraftGeomUtils

        base = None
        pl = obj.Placement

        # PropertySet support
        propSetPickedUuidPrev = self.ArchSkPropSetPickedUuid
        propSetListPrev = self.ArchSkPropSetListPrev
        propSetSelectedNamePrev = obj.ArchSketchPropertySet
        propSetSelectedNameCur = None
        propSetListCur = None
        if Draft.getType(obj.Base) == "ArchSketch":
            baseProxy = obj.Base.Proxy
            if hasattr(baseProxy, "getPropertySet"):
                # get full list of PropertySet
                propSetListCur = baseProxy.getPropertySet(obj.Base)
                # get updated name (if any) of the selected PropertySet
                propSetSelectedNameCur = baseProxy.getPropertySet(
                    obj.Base, propSetUuid=propSetPickedUuidPrev
                )
        if propSetSelectedNameCur:  # True if selection is not deleted
            if propSetListPrev != propSetListCur:
                obj.ArchSketchPropertySet = propSetListCur
                obj.ArchSketchPropertySet = propSetSelectedNameCur
                self.ArchSkPropSetListPrev = propSetListCur
            # elif propSetListPrev == propSetListCur:
            # pass  #nothing to do in this case
            # but if below, though (propSetListPrev == propSetListCur)
            elif propSetSelectedNamePrev != propSetSelectedNameCur:
                obj.ArchSketchPropertySet = propSetSelectedNameCur
        else:  # True if selection is deleted
            if propSetListCur:
                if propSetListPrev != propSetListCur:
                    obj.ArchSketchPropertySet = propSetListCur
                    obj.ArchSketchPropertySet = "Default"
                # else:  # Seems no need ...
                # obj.PropertySet = 'Default'

        extdata = self.getExtrusionData(obj)
        if extdata:
            bplates = extdata[0]
            extv = extdata[2].Rotation.multVec(extdata[1])
            if isinstance(bplates, list):
                shps = []
                # Test : if base is Sketch, then fuse all solid; otherwise, makeCompound
                sketchBaseToFuse = obj.Base.getLinkedObject().isDerivedFrom(
                    "Sketcher::SketchObject"
                )
                # but turn this off if we have layers, otherwise layers get merged
                if (
                    hasattr(obj, "Material")
                    and obj.Material
                    and hasattr(obj.Material, "Materials")
                    and obj.Material.Materials
                ):
                    sketchBaseToFuse = False
                for b in bplates:
                    b.Placement = extdata[2].multiply(b.Placement)
                    b = b.extrude(extv)

                    # See getExtrusionData() - not fusing baseplates there - fuse solids here
                    # Remarks - If solids are fused, but exportIFC.py use underlying baseplates w/o fuse, the result in ifc look slightly different from in FC.

                    if sketchBaseToFuse:
                        if shps:
                            shps = shps.fuse(b)  # shps.fuse(b)
                        else:
                            shps = b
                    else:
                        shps.append(b)
                    # TODO - To let user to select whether to fuse (slower) or to do a compound (faster) only ?

                if sketchBaseToFuse:
                    base = shps
                else:
                    base = Part.makeCompound(shps)
            else:
                bplates.Placement = extdata[2].multiply(bplates.Placement)
                base = bplates.extrude(extv)
        if obj.Base:
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape.isNull():
                    return
                if not obj.Base.Shape.isValid():
                    if not obj.Base.Shape.Solids:
                        # let pass invalid objects if they have solids...
                        return
                elif obj.Base.Shape.Solids:
                    base = Part.Shape(obj.Base.Shape)
                # blocks calculation
                elif hasattr(obj, "MakeBlocks") and hasattr(self, "basewires"):
                    if obj.MakeBlocks and self.basewires and extdata and obj.Width and obj.Height:
                        # print "calculating blocks"
                        if len(self.basewires) == 1:
                            blocks = []
                            n = FreeCAD.Vector(extv)
                            n.normalize()
                            cuts1 = []
                            cuts2 = []
                            if obj.BlockLength.Value:
                                for i in range(2):
                                    if i == 0:
                                        offset = obj.OffsetFirst.Value
                                    else:
                                        offset = obj.OffsetSecond.Value

                                    # only 1 wire (first) is supported
                                    # TODO - Can support multiple wires?

                                    # self.basewires was list of list of edges,
                                    # no matter Base is DWire, Sketch or else
                                    # See discussion - https://forum.freecad.org/viewtopic.php?t=86365
                                    baseEdges = self.basewires[0]

                                    for edge in baseEdges:
                                        while offset < (edge.Length - obj.Joint.Value):
                                            if offset:
                                                t = edge.tangentAt(offset)
                                                p = t.cross(n)
                                                p.multiply(1.1 * obj.Width.Value + obj.Offset.Value)
                                                p1 = edge.valueAt(offset).add(p)
                                                p2 = edge.valueAt(offset).add(p.negative())
                                                sh = Part.LineSegment(p1, p2).toShape()
                                                if obj.Joint.Value:
                                                    sh = sh.extrude(-t.multiply(obj.Joint.Value))
                                                sh = sh.extrude(n)
                                                if i == 0:
                                                    cuts1.append(sh)
                                                else:
                                                    cuts2.append(sh)
                                            offset += obj.BlockLength.Value + obj.Joint.Value
                                        offset -= edge.Length

                            if isinstance(bplates, list):
                                bplates = bplates[0]
                            if obj.BlockHeight.Value:
                                fsize = obj.BlockHeight.Value + obj.Joint.Value
                                bh = obj.BlockHeight.Value
                            else:
                                fsize = obj.Height.Value
                                bh = obj.Height.Value
                            bvec = FreeCAD.Vector(n)
                            bvec.multiply(bh)
                            svec = FreeCAD.Vector(n)
                            svec.multiply(fsize)
                            if cuts1:
                                plate1 = bplates.cut(cuts1).Faces
                            else:
                                plate1 = bplates.Faces
                            blocks1 = Part.makeCompound([f.extrude(bvec) for f in plate1])
                            if cuts2:
                                plate2 = bplates.cut(cuts2).Faces
                            else:
                                plate2 = bplates.Faces
                            blocks2 = Part.makeCompound([f.extrude(bvec) for f in plate2])
                            interval = extv.Length / (fsize)
                            entire = int(interval)
                            rest = interval - entire
                            for i in range(entire):
                                if i % 2:  # odd
                                    b = Part.Shape(blocks2)
                                else:
                                    b = Part.Shape(blocks1)
                                if i:
                                    t = FreeCAD.Vector(svec)
                                    t.multiply(i)
                                    b.translate(t)
                                blocks.append(b)
                            if rest:
                                rest = extv.Length - (entire * fsize)
                                rvec = FreeCAD.Vector(n)
                                rvec.multiply(rest)
                                if entire % 2:
                                    b = Part.makeCompound([f.extrude(rvec) for f in plate2])
                                else:
                                    b = Part.makeCompound([f.extrude(rvec) for f in plate1])
                                t = FreeCAD.Vector(svec)
                                t.multiply(entire)
                                b.translate(t)
                                blocks.append(b)
                            if blocks:
                                base = Part.makeCompound(blocks)

                        else:
                            FreeCAD.Console.PrintWarning(
                                translate("Arch", "Cannot compute blocks for wall")
                                + obj.Label
                                + "\n"
                            )

            elif obj.Base.isDerivedFrom("Mesh::Feature"):
                if obj.Base.Mesh.isSolid():
                    if obj.Base.Mesh.countComponents() == 1:
                        sh = ArchCommands.getShapeFromMesh(obj.Base.Mesh)
                        if sh.isClosed() and sh.isValid() and sh.Solids and (not sh.isNull()):
                            base = sh
                        else:
                            FreeCAD.Console.PrintWarning(
                                translate("Arch", "This mesh is an invalid solid") + "\n"
                            )
                            obj.Base.ViewObject.show()
        if not base:
            # FreeCAD.Console.PrintError(translate("Arch","Error: Invalid base object")+"\n")
            # return
            # walls can be made of only a series of additions and have no base shape
            base = Part.Shape()

        base = self.processSubShapes(obj, base, pl)

        self.applyShape(obj, base, pl)

        # count blocks
        if hasattr(obj, "MakeBlocks"):
            if obj.MakeBlocks:
                fvol = obj.BlockLength.Value * obj.BlockHeight.Value * obj.Width.Value
                if fvol:
                    # print("base volume:",fvol)
                    # for s in base.Solids:
                    # print(abs(s.Volume - fvol))
                    ents = [s for s in base.Solids if abs(s.Volume - fvol) < 1]
                    obj.CountEntire = len(ents)
                    obj.CountBroken = len(base.Solids) - len(ents)
                else:
                    obj.CountEntire = 0
                    obj.CountBroken = 0

        # set the length property
        if self.connectEdges:
            l = float(0)
            for e in self.connectEdges:
                l += e.Length
            l = l / 2
            if self.layersNum:
                l = l / self.layersNum
            if obj.Length.Value != l:
                obj.Length = l
                self.oldLength = (
                    None  # delete the stored value to prevent triggering base change below
                )

        # set the Area property
        obj.Area = obj.Length.Value * obj.Height.Value

    def onBeforeChange(self, obj, prop):
        """Method called before the object has a property changed.

        Specifically, this method is called before the value changes.

        If "Length" has changed, record the old length so that .onChanged() can
        be sure that the base needs to be changed.

        Also call ArchComponent.Component.onBeforeChange().

        Parameters
        ----------
        prop: string
            The name of the property that has changed.
        """

        if prop == "Length":
            self.oldLength = obj.Length.Value
        ArchComponent.Component.onBeforeChange(self, obj, prop)

    def onChanged(self, obj, prop):
        """Method called when the object has a property changed.

        If length has changed, extend the length of the Base object, if the
        Base object only has a single edge to extend.

        Also hide subobjects.

        Also call ArchComponent.Component.onChanged().

        Parameters
        ----------
        prop: string
            The name of the property that has changed.
        """

        if prop == "Length":
            if (
                obj.Base
                and obj.Length.Value
                and hasattr(self, "oldLength")
                and (self.oldLength is not None)
                and (self.oldLength != obj.Length.Value)
            ):
                if hasattr(obj.Base, "Shape"):
                    if len(obj.Base.Shape.Edges) == 1:
                        import DraftGeomUtils

                        e = obj.Base.Shape.Edges[0]
                        if DraftGeomUtils.geomType(e) == "Line":
                            if e.Length != obj.Length.Value:
                                v = e.Vertexes[-1].Point.sub(e.Vertexes[0].Point)
                                v.normalize()
                                v.multiply(obj.Length.Value)
                                p2 = e.Vertexes[0].Point.add(v)
                                if Draft.getType(obj.Base) == "Wire":
                                    # print "modifying p2"
                                    obj.Base.End = p2
                                elif Draft.getType(obj.Base) in [
                                    "Sketcher::SketchObject",
                                    "ArchSketch",
                                ]:
                                    # obj.Base.recompute() # Fix for the 'GeoId index out range' error. Not required in V1.1.
                                    obj.Base.moveGeometry(
                                        0, 2, obj.Base.Placement.inverse().multVec(p2)
                                    )
                                else:
                                    FreeCAD.Console.PrintError(
                                        translate(
                                            "Arch",
                                            "Error: Unable to modify the base object of this wall",
                                        )
                                        + "\n"
                                    )

        if prop == "ArchSketchPropertySet" and Draft.getType(obj.Base) == "ArchSketch":
            baseProxy = obj.Base.Proxy
            if hasattr(baseProxy, "getPropertySet"):
                uuid = baseProxy.getPropertySet(obj, propSetName=obj.ArchSketchPropertySet)
                self.ArchSkPropSetPickedUuid = uuid
        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj, "Width"):
                obj.setEditorMode("Width", ["ReadOnly"])
            if hasattr(obj, "Align"):
                obj.setEditorMode("Align", ["ReadOnly"])
            if hasattr(obj, "Offset"):
                obj.setEditorMode("Offset", ["ReadOnly"])
            if hasattr(obj, "OverrideWidth"):
                obj.setEditorMode("OverrideWidth", ["ReadOnly"])
            if hasattr(obj, "OverrideAlign"):
                obj.setEditorMode("OverrideAlign", ["ReadOnly"])
            if hasattr(obj, "OverrideOffset"):
                obj.setEditorMode("OverrideOffset", ["ReadOnly"])
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", ["ReadOnly"])
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", 0)
        else:
            if hasattr(obj, "Width"):
                obj.setEditorMode("Width", 0)
            if hasattr(obj, "Align"):
                obj.setEditorMode("Align", 0)
            if hasattr(obj, "Offset"):
                obj.setEditorMode("Offset", 0)
            if hasattr(obj, "OverrideWidth"):
                obj.setEditorMode("OverrideWidth", 0)
            if hasattr(obj, "OverrideAlign"):
                obj.setEditorMode("OverrideAlign", 0)
            if hasattr(obj, "OverrideOffset"):
                obj.setEditorMode("OverrideOffset", 0)
            if hasattr(obj, "ArchSketchEdges"):
                obj.setEditorMode("ArchSketchEdges", 0)
            if hasattr(obj, "ArchSketchPropertySet"):
                obj.setEditorMode("ArchSketchPropertySet", ["ReadOnly"])

        self.hideSubobjects(obj, prop)
        ArchComponent.Component.onChanged(self, obj, prop)

    def getFootprint(self, obj):
        """Get the faces that make up the base/foot of the wall.

        Returns
        -------
        list of <Part.Face>
            The faces that make up the foot of the wall.
        """

        faces = []
        if obj.Shape:
            for f in obj.Shape.Faces:
                if f.normalAt(0, 0).getAngle(FreeCAD.Vector(0, 0, -1)) < 0.01:
                    if abs(abs(f.CenterOfMass.z) - abs(obj.Shape.BoundBox.ZMin)) < 0.001:
                        faces.append(f)
        return faces

    def getExtrusionData(self, obj):
        """Get data needed to extrude the wall from a base object.

        take the Base object, and find a base face to extrude
        out, a vector to define the extrusion direction and distance.

        Rebase the base face to the (0,0,0) origin.

        Return the base face, rebased, with the extrusion vector, and the
        <Base.Placement> needed to return the face back to its original
        position.

        Returns
        -------
        tuple of (<Part.Face>, <Base.Vector>, <Base.Placement>)
            Tuple containing the base face, the vector for extrusion, and the
            placement needed to move the face back from the (0,0,0) origin.
        """

        import Part
        import DraftGeomUtils
        import ArchSketchObject

        propSetUuid = self.ArchSkPropSetPickedUuid

        # If ArchComponent.Component.getExtrusionData() can successfully get
        # extrusion data, just use that.
        data = ArchComponent.Component.getExtrusionData(self, obj)
        if data:
            if not isinstance(data[0], list):
                # multifuses not considered here
                return data
        length = obj.Length.Value
        # TODO currently layers were not supported when len(basewires) > 0	##( or 1 ? )
        width = 0
        # Get width of each edge segment from Base Objects if they store it
        # (Adding support in SketchFeaturePython, DWire...)
        widths = []  # [] or None are both False
        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj.Base, "Proxy"):  # TODO Any need to test ?
                if hasattr(obj.Base.Proxy, "getWidths"):
                    # Return a list of Width corresponding to indexes of sorted
                    # edges of Sketch.
                    widths = obj.Base.Proxy.getWidths(obj.Base, propSetUuid=propSetUuid)
        # Get width of each edge/wall segment from ArchWall.OverrideWidth if
        # Base Object does not provide it
        if not widths:
            if obj.OverrideWidth:
                if obj.Base and obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                    # If Base Object is ordinary Sketch (or when ArchSketch.getWidth() not implemented yet):-
                    # sort the width list in OverrrideWidth to correspond to indexes of sorted edges of Sketch
                    try:
                        import ArchSketchObject
                    except Exception:
                        print("ArchSketchObject add-on module is not installed yet")
                    try:
                        widths = ArchSketchObject.sortSketchWidth(
                            obj.Base, obj.OverrideWidth, obj.ArchSketchEdges
                        )
                    except Exception:
                        widths = obj.OverrideWidth
                else:
                    # If Base Object is not Sketch, but e.g. DWire, the width
                    # list in OverrrideWidth just correspond to sequential
                    # order of edges
                    widths = obj.OverrideWidth
            elif obj.Width:
                widths = [obj.Width.Value]
            else:
                # having no width is valid for walls so the user doesn't need to be warned
                # it just disables extrusions and return none
                # print ("Width & OverrideWidth & base.getWidths() should not be all 0 or None or [] empty list ")
                return None

        # Set 'default' width - for filling in any item in the list == 0 or None
        if obj.Width.Value:
            width = obj.Width.Value
        else:
            width = 200  # 'Default' width value

        # Get align of each edge segment from Base Objects if they store it.
        # (Adding support in SketchFeaturePython, DWire...)
        aligns = []
        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj.Base, "Proxy"):
                if hasattr(obj.Base.Proxy, "getAligns"):
                    # Return a list of Align corresponds to indexes of sorted
                    # edges of Sketch.
                    aligns = obj.Base.Proxy.getAligns(obj.Base, propSetUuid=propSetUuid)
        # Get align of each edge/wall segment from ArchWall.OverrideAlign if
        # Base Object does not provide it
        if not aligns:
            if obj.OverrideAlign:
                if obj.Base and obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                    # If Base Object is ordinary Sketch (or when
                    # ArchSketch.getAligns() not implemented yet):- sort the
                    # align list in OverrideAlign to correspond to indexes of
                    # sorted edges of Sketch
                    try:
                        import ArchSketchObject
                    except Exception:
                        print("ArchSketchObject add-on module is not installed yet")
                    try:
                        aligns = ArchSketchObject.sortSketchAlign(
                            obj.Base, obj.OverrideAlign, obj.ArchSketchEdges
                        )
                    except Exception:
                        aligns = obj.OverrideAlign
                else:
                    # If Base Object is not Sketch, but e.g. DWire, the align
                    # list in OverrideAlign just correspond to sequential order
                    # of edges
                    aligns = obj.OverrideAlign
            else:
                aligns = [obj.Align]

        # set 'default' align - for filling in any item in the list == 0 or None
        align = obj.Align  # or aligns[0]

        # Get offset of each edge segment from Base Objects if they store it
        # (Adding support in SketchFeaturePython, DWire...)
        offsets = []  # [] or None are both False
        if (
            hasattr(obj, "ArchSketchData")
            and obj.ArchSketchData
            and Draft.getType(obj.Base) == "ArchSketch"
        ):
            if hasattr(obj.Base, "Proxy"):
                if hasattr(obj.Base.Proxy, "getOffsets"):
                    # Return a list of Offset corresponding to indexes of sorted
                    # edges of Sketch.
                    offsets = obj.Base.Proxy.getOffsets(obj.Base, propSetUuid=propSetUuid)
        # Get offset of each edge/wall segment from ArchWall.OverrideOffset if
        # Base Object does not provide it
        if not offsets:
            if obj.OverrideOffset:
                if obj.Base and obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                    # If Base Object is ordinary Sketch (or when ArchSketch.getOffsets() not implemented yet):-
                    # sort the offset list in OverrideOffset to correspond to indexes of sorted edges of Sketch
                    if hasattr(ArchSketchObject, "sortSketchOffset"):
                        offsets = ArchSketchObject.sortSketchOffset(
                            obj.Base, obj.OverrideOffset, obj.ArchSketchEdges
                        )
                    else:
                        offsets = obj.OverrideOffset
                else:
                    # If Base Object is not Sketch, but e.g. DWire, the width
                    # list in OverrrideWidth just correspond to sequential
                    # order of edges
                    offsets = obj.OverrideOffset
            elif obj.Offset:
                offsets = [obj.Offset.Value]

        # Set 'default' offset - for filling in any item in the list == 0 or None
        offset = obj.Offset.Value  # could be 0

        height = obj.Height.Value
        if not height:
            height = self.getParentHeight(obj)
        if not height:
            return None
        if obj.Normal == Vector(0, 0, 0):
            if obj.Base and hasattr(obj.Base, "Shape"):
                normal = DraftGeomUtils.get_shape_normal(obj.Base.Shape)
                if normal is None:
                    normal = Vector(0, 0, 1)
            else:
                normal = Vector(0, 0, 1)
        else:
            normal = Vector(obj.Normal)
        base = None
        placement = None
        self.basewires = None

        # build wall layers
        layers = []
        if hasattr(obj, "Material"):
            if obj.Material:
                if hasattr(obj.Material, "Materials"):
                    thicknesses = [abs(t) for t in obj.Material.Thicknesses]
                    # multimaterials
                    varwidth = 0
                    restwidth = width - sum(thicknesses)
                    if restwidth > 0:
                        varwidth = [t for t in thicknesses if t == 0]
                        if varwidth:
                            varwidth = restwidth / len(varwidth)
                    for t in obj.Material.Thicknesses:
                        if t:
                            layers.append(t)
                        elif varwidth:
                            layers.append(varwidth)

        # Check if there is obj.Base and its validity to proceed
        if self.ensureBase(obj):
            if hasattr(obj.Base, "Shape"):
                if obj.Base.Shape:
                    if obj.Base.Shape.Solids:
                        return None

                    # If the user has defined a specific face of the Base
                    # object to build the wall from, extrude from that face,
                    # and return the extrusion moved to (0,0,0), normal of the
                    # face, and placement to move the extrusion back to its
                    # original position.
                    elif obj.Face > 0:
                        if len(obj.Base.Shape.Faces) >= obj.Face:
                            face = obj.Base.Shape.Faces[obj.Face - 1]
                            if obj.Normal != Vector(0, 0, 0):
                                normal = face.normalAt(0, 0)
                            if normal.getAngle(Vector(0, 0, 1)) > math.pi / 4:
                                normal.multiply(width)
                                base = face.extrude(normal)
                                if obj.Align == "Center":
                                    base.translate(normal.negative().multiply(0.5))
                                elif obj.Align == "Right":
                                    base.translate(normal.negative())
                            else:
                                normal.multiply(height)
                                base = face.extrude(normal)
                            base, placement = self.rebase(base)
                            return (base, normal, placement)

                    # If the Base has faces, but no specific one has been
                    # selected, rebase the faces and continue.
                    elif obj.Base.Shape.Faces:
                        if not DraftGeomUtils.isCoplanar(obj.Base.Shape.Faces):
                            return None
                        else:
                            base, placement = self.rebase(obj.Base.Shape)

                    elif (
                        hasattr(obj.Base, "Proxy")
                        and obj.ArchSketchData
                        and hasattr(obj.Base.Proxy, "getWallBaseShapeEdgesInfo")
                    ):
                        wallBaseShapeEdgesInfo = obj.Base.Proxy.getWallBaseShapeEdgesInfo(
                            obj.Base, propSetUuid=propSetUuid
                        )
                        # get wall edges (not wires); use original edges if getWallBaseShapeEdgesInfo() provided none
                        if wallBaseShapeEdgesInfo:
                            self.basewires = wallBaseShapeEdgesInfo.get(
                                "wallAxis"
                            )  # 'wallEdges'  # widths, aligns, offsets?

                    # Sort Sketch edges consistently with below procedures
                    # without using Sketch.Shape.Edges - found the latter order
                    # in some corner case != getSortedClusters()
                    elif obj.Base.isDerivedFrom("Sketcher::SketchObject"):
                        self.basewires = []
                        skGeom = obj.Base.GeometryFacadeList
                        skGeomEdges = []
                        skPlacement = obj.Base.Placement  # Get Sketch's placement to restore later
                        # Get ArchSketch edges to construct ArchWall
                        # No need to test obj.ArchSketchData ...
                        for ig, geom in enumerate(skGeom):
                            # Construction mode edges should be ignored if
                            # ArchSketchEdges, otherwise, ArchSketchEdges data
                            # needs to take out those in Construction before
                            # using as parameters.
                            if (not obj.ArchSketchEdges and not geom.Construction) or str(
                                ig
                            ) in obj.ArchSketchEdges:
                                # support Line, Arc, Circle, Ellipse for Sketch
                                # as Base at the moment
                                if isinstance(
                                    geom.Geometry,
                                    (Part.LineSegment, Part.Circle, Part.ArcOfCircle, Part.Ellipse),
                                ):
                                    skGeomEdgesI = geom.Geometry.toShape()

                                    skGeomEdges.append(skGeomEdgesI)
                        for cluster in Part.getSortedClusters(skGeomEdges):
                            clusterTransformed = []
                            for edge in cluster:
                                # TODO 2023.11.26: Multiplication order should be switched?
                                # So far 'no problem' as 'edge.placement' is always '0,0,0' ?
                                edge.Placement = edge.Placement.multiply(
                                    skPlacement
                                )  ## TODO add attribute to skip Transform...

                                clusterTransformed.append(edge)
                            # Only use cluster of edges rather than turning into wire
                            self.basewires.append(clusterTransformed)

                        # Use Sketch's Normal for all edges/wires generated
                        # from sketch for consistency. Discussion on checking
                        # normal of sketch.Placement vs
                        # sketch.getGlobalPlacement() -
                        # https://forum.freecad.org/viewtopic.php?f=22&t=39341&p=334275#p334275
                        # normal = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(0,0,1))
                        normal = obj.Base.getGlobalPlacement().Rotation.multVec(
                            FreeCAD.Vector(0, 0, 1)
                        )

                    else:  # For all objects except Sketch, single edge or more
                        # See discussion - https://forum.freecad.org/viewtopic.php?t=86365
                        # See discussion - https://forum.freecad.org/viewtopic.php?t=82207&start=10
                        # self.basewires = obj.Base.Shape.Wires
                        #
                        # Now, adopt approach same as for Sketch
                        self.basewires = []
                        clusters = Part.getSortedClusters(obj.Base.Shape.Edges)
                        self.basewires = clusters
                        # Previously :
                        # Found case that after sorting below, direction of
                        # edges sorted are not as 'expected' thus resulted in
                        # bug - e.g. a Dwire with edges/vertexes in clockwise
                        # order, 1st vertex is Forward as expected.  After
                        # sorting below, edges sorted still in clockwise order
                        # - no problem, but 1st vertex of each edge become
                        # Reverse rather than Forward.

                        # See FC discussion -
                        # https://forum.freecad.org/viewtopic.php?f=23&t=48275&p=413745#p413745

                        # self.basewires = []
                        # for cluster in Part.getSortedClusters(obj.Base.Shape.Edges):
                        #    for c in Part.sortEdges(cluster):
                        #        self.basewires.append(Part.Wire(c))
                        # if not sketch, e.g. Dwire, can have wire which is 3d
                        # so not on the placement's working plane - below
                        # applied to Sketch not applicable here
                        # normal = obj.Base.getGlobalPlacement().Rotation.multVec(FreeCAD.Vector(0,0,1))
                        # normal = obj.Base.Placement.Rotation.multVec(FreeCAD.Vector(0,0,1))

                    if self.basewires:
                        if (len(self.basewires) == 1) and layers:
                            self.basewires = [self.basewires[0] for l in layers]
                            self.layersNum = len(layers)
                        else:
                            self.layersNum = 0
                        layeroffset = 0
                        baseface = None
                        self.connectEdges = []
                        for i, wire in enumerate(self.basewires):

                            # Check number of edges per 'wire' and get the 1st edge
                            if isinstance(wire, Part.Wire):
                                edgeNum = len(wire.Edges)
                                e = wire.Edges[0]
                            elif isinstance(wire[0], Part.Edge):
                                edgeNum = len(wire)
                                e = wire[0]

                            for n in range(
                                0, edgeNum, 1
                            ):  # why these not work - range(edgeNum), range(0,edgeNum) ...

                                # Fill the aligns list with ArchWall's default
                                # align entry and with same number of items as
                                # number of edges
                                try:
                                    if aligns[n] not in ["Left", "Right", "Center"]:
                                        aligns[n] = align
                                except Exception:
                                    aligns.append(align)

                                # Fill the widths List with ArchWall's default
                                # width entry and with same number of items as
                                # number of edges
                                try:
                                    if not widths[n]:
                                        widths[n] = width
                                except Exception:
                                    widths.append(width)
                                # Fill the offsets List with ArchWall's default
                                # offset entry and with same number of items as
                                # number of edges
                                try:
                                    if not offsets[n]:
                                        offsets[n] = offset
                                except Exception:
                                    offsets.append(offset)

                            # Get a direction vector orthogonal to both the
                            # normal of the face/sketch and the direction the
                            # wire was drawn in. IE: along the width direction
                            # of the wall.
                            if isinstance(e.Curve, (Part.Circle, Part.Ellipse)):
                                dvec = e.Vertexes[0].Point.sub(e.Curve.Center)
                            else:
                                dvec = DraftGeomUtils.vec(e).cross(normal)

                            if not DraftVecUtils.isNull(dvec):
                                dvec.normalize()
                            face = None

                            curAligns = aligns[0]
                            # off = obj.Offset.Value  # off is no longer used

                            if curAligns == "Left":

                                if layers:
                                    curWidth = []
                                    for n in range(edgeNum):
                                        curWidth.append(abs(layers[i]))
                                    # off = off+layeroffset  # off is no longer used
                                    offsets = [x + layeroffset for x in offsets]
                                    dvec.multiply(curWidth[0])
                                    layeroffset += abs(curWidth[0])
                                else:
                                    curWidth = widths
                                    dvec.multiply(width)

                                # Now DraftGeomUtils.offsetWire() support
                                # similar effect as ArchWall Offset
                                #
                                # if off:
                                #    dvec2 = DraftVecUtils.scaleTo(dvec,off)
                                #    wire = DraftGeomUtils.offsetWire(wire,dvec2)

                                # Get the 'offseted' wire taking into account
                                # of Width and Align of each edge, and overall
                                # Offset
                                wNe2 = DraftGeomUtils.offsetWire(
                                    wire,
                                    dvec,
                                    bind=False,
                                    occ=False,
                                    widthList=curWidth,
                                    offsetMode=None,
                                    alignList=aligns,
                                    normal=normal,
                                    basewireOffset=offsets,
                                    wireNedge=True,
                                )
                                # Get the 'base' wire taking into account of
                                # width and align of each edge
                                wNe1 = DraftGeomUtils.offsetWire(
                                    wire,
                                    dvec,
                                    bind=False,
                                    occ=False,
                                    widthList=curWidth,
                                    offsetMode="BasewireMode",
                                    alignList=aligns,
                                    normal=normal,
                                    basewireOffset=offsets,
                                    wireNedge=True,
                                )
                            elif curAligns == "Right":
                                dvec = dvec.negative()

                                if layers:
                                    curWidth = []
                                    for n in range(edgeNum):
                                        curWidth.append(abs(layers[i]))
                                    # off = off+layeroffset  # off is no longer used
                                    offsets = [x + layeroffset for x in offsets]
                                    dvec.multiply(curWidth[0])
                                    layeroffset += abs(curWidth[0])
                                else:
                                    curWidth = widths
                                    dvec.multiply(width)

                                # Now DraftGeomUtils.offsetWire() support similar effect as ArchWall Offset
                                #
                                # if off:
                                #    dvec2 = DraftVecUtils.scaleTo(dvec,off)
                                #    wire = DraftGeomUtils.offsetWire(wire,dvec2)
                                wNe2 = DraftGeomUtils.offsetWire(
                                    wire,
                                    dvec,
                                    bind=False,
                                    occ=False,
                                    widthList=curWidth,
                                    offsetMode=None,
                                    alignList=aligns,
                                    normal=normal,
                                    basewireOffset=offsets,
                                    wireNedge=True,
                                )
                                wNe1 = DraftGeomUtils.offsetWire(
                                    wire,
                                    dvec,
                                    bind=False,
                                    occ=False,
                                    widthList=curWidth,
                                    offsetMode="BasewireMode",
                                    alignList=aligns,
                                    normal=normal,
                                    basewireOffset=offsets,
                                    wireNedge=True,
                                )
                            elif curAligns == "Center":
                                if layers:
                                    # TODO Current, the order of layers follow
                                    # "Right" align.  Option for the order of
                                    # layers follow "Left" align should be
                                    # provided for users.
                                    dvec = dvec.negative()
                                    totalwidth = sum([abs(l) for l in layers])
                                    off = totalwidth / 2 - layeroffset
                                    # TODO To consider offset per edge?
                                    #
                                    # Offset follows direction of align "Right".
                                    # Needs to be reversed in this case.
                                    off = -off
                                    # d1 = Vector(dvec).multiply(off)
                                    curWidth = []
                                    alignListC = []
                                    offsetListC = []
                                    for n in range(edgeNum):
                                        curWidth.append(abs(layers[i]))
                                        alignListC.append("Right")  # ("Left")
                                        offsetListC.append(off)
                                    # wNe1 = DraftGeomUtils.offsetWire(wire, d1, wireNedge=True)
                                    # See https://github.com/FreeCAD/FreeCAD/issues/25485#issuecomment-3566734050
                                    # d1 may be Vector (0,0,0), offsetWire()
                                    # in draftgeoutils\offsets.py:-
                                    # v1 = App.Vector(dvec).normalize() return
                                    # error.  Provide widthList, alignList etc.
                                    # so no need to run above code to deduce
                                    # v1 in offsetWire()
                                    wNe1 = DraftGeomUtils.offsetWire(
                                        wire,
                                        dvec,
                                        widthList=curWidth,
                                        offsetMode="BasewireMode",
                                        alignList=alignListC,
                                        normal=normal,
                                        basewireOffset=offsetListC,
                                        wireNedge=True,
                                    )
                                    wNe2 = DraftGeomUtils.offsetWire(
                                        wire,
                                        dvec,
                                        widthList=curWidth,
                                        offsetMode=None,
                                        alignList=alignListC,
                                        normal=normal,
                                        basewireOffset=offsetListC,
                                        wireNedge=True,
                                    )
                                    layeroffset += abs(curWidth[0])
                                else:
                                    dvec.multiply(width)
                                    wNe2 = DraftGeomUtils.offsetWire(
                                        wire,
                                        dvec,
                                        bind=False,
                                        occ=False,
                                        widthList=widths,
                                        offsetMode=None,
                                        alignList=aligns,
                                        normal=normal,
                                        basewireOffset=offsets,
                                        wireNedge=True,
                                    )
                                    wNe1 = DraftGeomUtils.offsetWire(
                                        wire,
                                        dvec,
                                        bind=False,
                                        occ=False,
                                        widthList=widths,
                                        offsetMode="BasewireMode",
                                        alignList=aligns,
                                        normal=normal,
                                        basewireOffset=offsets,
                                        wireNedge=True,
                                    )
                            w2 = wNe2[0]
                            w1 = wNe1[0]
                            face = DraftGeomUtils.bind(w1, w2, per_segment=True)
                            cEdgesF2 = wNe2[1]
                            cEdges2 = wNe2[2]
                            oEdges2 = wNe2[3]
                            cEdgesF1 = wNe1[1]
                            cEdges1 = wNe1[2]
                            oEdges1 = wNe1[3]
                            self.connectEdges.extend(cEdges2)
                            self.connectEdges.extend(cEdges1)

                            del widths[0:edgeNum]
                            del aligns[0:edgeNum]
                            del offsets[0:edgeNum]

                            if face:

                                if layers and (layers[i] < 0):
                                    # layers with negative values are not drawn
                                    continue

                                if baseface:

                                    # To allow exportIFC.py to work properly on
                                    # sketch, which use only 1st face / wire,
                                    # do not fuse baseface here So for a sketch
                                    # with multiple wires, each returns
                                    # individual face (rather than fusing
                                    # together) for exportIFC.py to work
                                    # properly
                                    # "ArchWall - Based on Sketch Issues" - https://forum.freecad.org/viewtopic.php?f=39&t=31235

                                    # "Bug #2408: [PartDesign] .fuse is splitting edges it should not"
                                    # - https://forum.freecad.org/viewtopic.php?f=10&t=20349&p=346237#p346237
                                    # - bugtracker - https://freecad.org/tracker/view.php?id=2408

                                    # Try Part.Shell before removeSplitter
                                    # - https://forum.freecad.org/viewtopic.php?f=10&t=20349&start=10
                                    # - 1st finding : if a rectangle + 1 line, can't removesSplitter properly...
                                    # - 2nd finding : if 2 faces do not touch, can't form a shell; then, subsequently for remaining faces even though touch each faces, can't form a shell

                                    baseface.append(face)
                                    # The above make Refine methods below (in else) useless, regardless removeSpitters yet to be improved for cases do not work well
                                    """  Whether layers or not, all baseface.append(face) """

                                else:
                                    baseface = [face]

                                    """  Whether layers or not, all baseface = [face] """

                        if baseface:
                            base, placement = self.rebase(baseface)

        # Build Wall if there is no obj.Base or even obj.Base is not valid
        else:
            if layers:
                totalwidth = sum([abs(l) for l in layers])
                offset = 0
                base = []
                for l in layers:
                    if l > 0:
                        l2 = length / 2 or 0.5
                        w1 = -totalwidth / 2 + offset
                        w2 = w1 + l
                        v1 = Vector(-l2, w1, 0)
                        v2 = Vector(l2, w1, 0)
                        v3 = Vector(l2, w2, 0)
                        v4 = Vector(-l2, w2, 0)
                        base.append(Part.Face(Part.makePolygon([v1, v2, v3, v4, v1])))
                    offset += abs(l)
            else:
                l2 = length / 2 or 0.5
                w2 = width / 2 or 0.5
                v1 = Vector(-l2, -w2, 0)
                v2 = Vector(l2, -w2, 0)
                v3 = Vector(l2, w2, 0)
                v4 = Vector(-l2, w2, 0)
                base = Part.Face(Part.makePolygon([v1, v2, v3, v4, v1]))
            placement = FreeCAD.Placement()
        if base and placement:
            normal.normalize()
            extrusion = normal.multiply(height)
            if placement.Rotation.Angle > 0:
                extrusion = placement.inverse().Rotation.multVec(extrusion)
            return (base, extrusion, placement)
        return None


class _ViewProviderWall(ArchComponent.ViewProviderComponent):
    """The view provider for the wall object.

    Parameters
    ----------
    vobj: <Gui.ViewProviderDocumentObject>
        The view provider to turn into a wall view provider.
    """

    def __init__(self, vobj):
        ArchComponent.ViewProviderComponent.__init__(self, vobj)
        vobj.ShapeColor = ArchCommands.getDefaultColor("Wall")

    def getIcon(self):
        """Return the path to the appropriate icon.

        If a clone, return the cloned wall icon path. Otherwise return the
        Arch wall icon.

        Returns
        -------
        str
            Path to the appropriate icon .svg file.
        """

        import Arch_rc

        if hasattr(self, "Object"):
            if self.Object.CloneOf:
                return ":/icons/Arch_Wall_Clone.svg"
            elif (not self.Object.Base) and self.Object.Additions:
                return ":/icons/Arch_Wall_Tree_Assembly.svg"
        return ":/icons/Arch_Wall_Tree.svg"

    def attach(self, vobj):
        """Add display modes' data to the coin scenegraph.

        Add each display mode as a coin node, whose parent is this view
        provider.

        Each display mode's node includes the data needed to display the object
        in that mode. This might include colors of faces, or the draw style of
        lines. This data is stored as additional coin nodes which are children
        of the display mode node.

        Add the textures used in the Footprint display mode.
        """

        self.Object = vobj.Object
        from pivy import coin

        tex = coin.SoTexture2()
        image = Draft.loadTexture(Draft.svgpatterns()["simple"][1], 128)
        if not image is None:
            tex.image = image
        texcoords = coin.SoTextureCoordinatePlane()
        s = params.get_param_arch("patternScale")
        texcoords.directionS.setValue(s, 0, 0)
        texcoords.directionT.setValue(0, s, 0)
        self.fcoords = coin.SoCoordinate3()
        self.fset = coin.SoIndexedFaceSet()
        sep = coin.SoSeparator()
        sep.addChild(tex)
        sep.addChild(texcoords)
        sep.addChild(self.fcoords)
        sep.addChild(self.fset)
        vobj.RootNode.addChild(sep)
        ArchComponent.ViewProviderComponent.attach(self, vobj)

    def updateData(self, obj, prop):
        """Method called when the host object has a property changed.

        If the host object's Placement, Shape, or Material has changed, and the
        host object has a Material assigned, give the shape the color and
        transparency of the Material.

        Parameters
        ----------
        obj: <App::FeaturePython>
            The host object that has changed.
        prop: string
            The name of the property that has changed.
        """

        if prop in ["Placement", "Shape", "Material"]:
            if obj.ViewObject.DisplayMode == "Footprint":
                obj.ViewObject.Proxy.setDisplayMode("Footprint")
            if hasattr(obj, "Material"):
                if obj.Material and obj.Shape:
                    if hasattr(obj.Material, "Materials"):
                        activematerials = [
                            obj.Material.Materials[i]
                            for i in range(len(obj.Material.Materials))
                            if obj.Material.Thicknesses[i] >= 0
                        ]
                        if len(activematerials) == len(obj.Shape.Solids):
                            cols = []
                            for i, mat in enumerate(activematerials):
                                c = obj.ViewObject.ShapeColor
                                c = (c[0], c[1], c[2], 1.0 - obj.ViewObject.Transparency / 100.0)
                                if "DiffuseColor" in mat.Material:
                                    if "(" in mat.Material["DiffuseColor"]:
                                        c = tuple(
                                            [
                                                float(f)
                                                for f in mat.Material["DiffuseColor"]
                                                .strip("()")
                                                .split(",")
                                            ]
                                        )
                                if "Transparency" in mat.Material:
                                    c = (
                                        c[0],
                                        c[1],
                                        c[2],
                                        1.0 - float(mat.Material["Transparency"]),
                                    )
                                cols.extend([c for j in range(len(obj.Shape.Solids[i].Faces))])
                            obj.ViewObject.DiffuseColor = cols
        ArchComponent.ViewProviderComponent.updateData(self, obj, prop)
        if len(obj.ViewObject.DiffuseColor) > 1:
            # force-reset colors if changed
            obj.ViewObject.DiffuseColor = obj.ViewObject.DiffuseColor

    def getDisplayModes(self, vobj):
        """Define the display modes unique to the Arch Wall.

        Define mode Footprint, which only displays the footprint of the wall.
        Also add the display modes of the Arch Component.

        Returns
        -------
        list of str
            List containing the names of the new display modes.
        """

        modes = ArchComponent.ViewProviderComponent.getDisplayModes(self, vobj) + ["Footprint"]
        return modes

    def setDisplayMode(self, mode):
        """Method called when the display mode changes.

        Called when the display mode changes, this method can be used to set
        data that wasn't available when .attach() was called.

        When Footprint is set as display mode, find the faces that make up the
        footprint of the wall, and give them a lined texture. Then display
        the wall as a wireframe.

        Then pass the displaymode onto Arch Component's .setDisplayMode().

        Parameters
        ----------
        mode: str
            The name of the display mode the view provider has switched to.

        Returns
        -------
        str:
            The name of the display mode the view provider has switched to.
        """

        self.fset.coordIndex.deleteValues(0)
        self.fcoords.point.deleteValues(0)
        if mode == "Footprint":
            if hasattr(self, "Object"):
                faces = self.Object.Proxy.getFootprint(self.Object)
                if faces:
                    verts = []
                    fdata = []
                    idx = 0
                    for face in faces:
                        tri = face.tessellate(1)
                        for v in tri[0]:
                            verts.append([v.x, v.y, v.z])
                        for f in tri[1]:
                            fdata.extend([f[0] + idx, f[1] + idx, f[2] + idx, -1])
                        idx += len(tri[0])
                    self.fcoords.point.setValues(verts)
                    self.fset.coordIndex.setValues(0, len(fdata), fdata)
            return "Wireframe"
        return ArchComponent.ViewProviderComponent.setDisplayMode(self, mode)

    def setupContextMenu(self, vobj, menu):

        if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
            return

        super().contextMenuAddEdit(menu)

        actionFlipDirection = QtGui.QAction(
            QtGui.QIcon(":/icons/Arch_Wall_Tree.svg"), translate("Arch", "Flip Direction"), menu
        )
        QtCore.QObject.connect(
            actionFlipDirection, QtCore.SIGNAL("triggered()"), self.flipDirection
        )
        menu.addAction(actionFlipDirection)

        super().contextMenuAddToggleSubcomponents(menu)

    def flipDirection(self):

        if hasattr(self, "Object") and self.Object:
            obj = self.Object
            if obj.Align == "Left":
                obj.Align = "Right"
                FreeCAD.ActiveDocument.recompute()
            elif obj.Align == "Right":
                obj.Align = "Left"
                FreeCAD.ActiveDocument.recompute()



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
import FreeCADGui
from PySide import QtGui

translate = FreeCAD.Qt.translate

PROP_TYPE_QTYES = ["App::PropertyDistance", "App::PropertyAngle"]
PROP_TYPE_NUMERIC = PROP_TYPE_QTYES + [
    "App::PropertyFloat",
    "App::PropertyInteger",
    "App::PropertyLength",
    "App::PropertyPercent",
]


class SimpleEditPanel:
    """A simple property edit panel for a FreeCAD object

    Attributes:
        _transaction_name: Name to use for the undo transaction used in this panel.
        _ui_file: Path to the Qt UI file to use when creating this panel. Must be
            overridden by the subclass.
    """

    _transaction_name = "Property Edit"
    _ui_file = ""

    def __init__(self, obj, view):
        self.obj = obj
        self.viewProvider = view
        FreeCAD.ActiveDocument.openTransaction(self._transaction_name)
        self.form = FreeCADGui.PySideUic.loadUi(self._ui_file)
        self._fc = {}
        self.setupUi()

    def getFields(self):
        for prop_name, (get_field, set_field) in self._fc.items():
            val = get_field()
            if isinstance(getattr(self.obj, prop_name), int):
                val = int(val)
            setattr(self.obj, prop_name, val)

    def setFields(self):
        for prop_name, (get_field, set_field) in self._fc.items():
            set_field(getattr(self.obj, prop_name))

    def connectWidget(self, prop_name, widget, custom_lbls=None):
        """Connect a widget to a proxy object property

        This registers a connection between a UI widget and an object property,
        performing a series of convenience functions in the process, namely:
          * It copies the tooltip from the object property, with proper handling
            of translation.
          * If it is an enum property, it loads the translated choices from the
            property into the widget. Optionally the caller can override some or
            all of these labels using the `custom_lbls` dictionary.
          * It registers type conversion functions so when the form is applied or
            accepted the data is converted and stored/retrieved without manual
            intervention.
        """
        if custom_lbls is None:
            custom_lbls = {}
        prop_type = self.obj.getTypeIdOfProperty(prop_name)
        widget_type = type(widget).__name__
        if prop_type == "App::PropertyEnumeration" and widget_type == "QComboBox":
            enum = self.obj.getEnumerationsOfProperty(prop_name)
            # Populate the combo box with the enumeration elements, use the form context for translation
            elements = [
                translate(self.form.objectName(), custom_lbls.get(itm, itm)) for itm in enum
            ]
            widget.clear()
            widget.addItems(elements)

            def _getter():
                return enum[widget.currentIndex()]

            def _setter(val):
                widget.setCurrentIndex(enum.index(val))

            self._fc[prop_name] = _getter, _setter
        elif prop_type == "App::PropertyBool" and widget_type == "QCheckBox":
            self._fc[prop_name] = widget.isChecked, widget.setChecked
        elif prop_type in PROP_TYPE_NUMERIC and widget_type == "QDoubleSpinBox":
            self._fc[prop_name] = widget.value, widget.setValue
        elif prop_type in PROP_TYPE_QTYES and widget_type == "QLineEdit":
            self._fc[prop_name] = widget.text, lambda v: widget.setText(v.UserString)
        else:
            raise ValueError(
                f"Unsupported connection between '{prop_type}' property and '{widget_type}' widget"
            )
        # Set the tooltip to the one corresponding to the property.
        widget.setToolTip(
            translate("App::Property", self.obj.getDocumentationOfProperty(prop_name))
        )

    def getStandardButtons(self):
        return (
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Apply | QtGui.QDialogButtonBox.Cancel
        )

    def clicked(self, button):
        # callback for standard buttons
        if button == QtGui.QDialogButtonBox.Apply:
            self.updateModel()
            FreeCAD.ActiveDocument.recompute()
        if button == QtGui.QDialogButtonBox.Cancel:
            self.abort()

    def abort(self):
        FreeCAD.ActiveDocument.abortTransaction()
        self.cleanup(True)

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.getFields()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def cleanup(self, gui):
        self.viewProvider.clearTaskPanel()
        if gui:
            FreeCADGui.Control.closeDialog()
            FreeCAD.ActiveDocument.recompute()

    def updateModel(self):
        self.getFields()
        self.obj.Proxy.execute(self.obj)
        FreeCAD.ActiveDocument.recompute()

    def open(self):
        pass



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Defines a class derived from QWidget for displaying the bar at the top of the addons list."""

from enum import IntEnum
import os

from PySideWrapper import QtCore, QtGui, QtWidgets
from .addonmanager_widget_view_selector import WidgetViewSelector
from .addonmanager_widget_filter_selector import WidgetFilterSelector
from .addonmanager_widget_search import WidgetSearch

translate = QtCore.QCoreApplication.translate


class SortOptions(IntEnum):
    _SortRoleOffset = 100
    Alphabetical = QtCore.Qt.UserRole + _SortRoleOffset + 0
    LastUpdated = QtCore.Qt.UserRole + _SortRoleOffset + 1
    DateAdded = QtCore.Qt.UserRole + _SortRoleOffset + 2
    Stars = QtCore.Qt.UserRole + _SortRoleOffset + 3
    Score = QtCore.Qt.UserRole + _SortRoleOffset + 4


default_sort_order = {
    SortOptions.Alphabetical: QtCore.Qt.AscendingOrder,
    SortOptions.LastUpdated: QtCore.Qt.DescendingOrder,
    SortOptions.DateAdded: QtCore.Qt.DescendingOrder,
    SortOptions.Stars: QtCore.Qt.DescendingOrder,
    SortOptions.Score: QtCore.Qt.DescendingOrder,
}


class WidgetViewControlBar(QtWidgets.QWidget):
    """A bar containing a view selection widget, a filter widget, and a search widget"""

    view_changed = QtCore.Signal(int)
    filter_changed = QtCore.Signal(object)
    search_changed = QtCore.Signal(str)
    sort_changed = QtCore.Signal(int)
    sort_order_changed = QtCore.Signal(QtCore.Qt.SortOrder)

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.has_rankings = False
        self._setup_ui()
        self._setup_connections()
        self.retranslateUi(None)
        self.sort_order = QtCore.Qt.AscendingOrder
        self._set_sort_order_icon()

    def _setup_ui(self):
        self.horizontal_layout = QtWidgets.QHBoxLayout()
        self.horizontal_layout.setContentsMargins(0, 0, 0, 0)
        self.view_selector = WidgetViewSelector(self)
        self.filter_selector = WidgetFilterSelector(self)
        self.sort_selector = QtWidgets.QComboBox(self)
        self.sort_order_button = QtWidgets.QToolButton(self)
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        self.sort_order_button.setIcon(
            QtGui.QIcon.fromTheme(
                "ascending", QtGui.QIcon(os.path.join(icon_path, "sort_ascending.svg"))
            )
        )
        self.search = WidgetSearch(self)
        self.horizontal_layout.addWidget(self.view_selector)
        self.horizontal_layout.addWidget(self.filter_selector)
        self.horizontal_layout.addWidget(self.sort_selector)
        self.horizontal_layout.addWidget(self.sort_order_button)
        self.horizontal_layout.addWidget(self.search)
        self.setLayout(self.horizontal_layout)

    def _sort_order_clicked(self):
        if self.sort_order == QtCore.Qt.AscendingOrder:
            self.set_sort_order(QtCore.Qt.DescendingOrder)
        else:
            self.set_sort_order(QtCore.Qt.AscendingOrder)
        self.sort_order_changed.emit(self.sort_order)

    def set_sort_order(self, order: QtCore.Qt.SortOrder) -> None:
        self.sort_order = order
        self._set_sort_order_icon()

    def _set_sort_order_icon(self):
        icon_path = os.path.join(os.path.dirname(__file__), "..", "Resources", "icons")
        if self.sort_order == QtCore.Qt.AscendingOrder:
            self.sort_order_button.setIcon(
                QtGui.QIcon.fromTheme(
                    "view-sort-ascending",
                    QtGui.QIcon(os.path.join(icon_path, "sort_ascending.svg")),
                )
            )
        else:
            self.sort_order_button.setIcon(
                QtGui.QIcon.fromTheme(
                    "view-sort-descending",
                    QtGui.QIcon(os.path.join(icon_path, "sort_descending.svg")),
                )
            )

    def set_rankings_available(self, rankings_available: bool) -> None:
        self.has_rankings = rankings_available
        self.retranslateUi(None)

    def _setup_connections(self):
        self.view_selector.view_changed.connect(self.view_changed.emit)
        self.filter_selector.filter_changed.connect(self.filter_changed.emit)
        self.search.search_changed.connect(self.search_changed.emit)
        self.sort_selector.currentIndexChanged.connect(self._sort_changed)
        self.sort_order_button.clicked.connect(self._sort_order_clicked)

    def _sort_changed(self, index: int):
        sort_role = self.sort_selector.itemData(index)
        if sort_role is None:
            sort_role = SortOptions.Alphabetical
        self.set_sort_order(default_sort_order[sort_role])
        self.sort_changed.emit(sort_role)
        self.sort_order_changed.emit(self.sort_order)

    def retranslateUi(self, _=None):
        self.sort_selector.clear()
        self.sort_selector.addItem(
            translate("AddonsInstaller", "Alphabetical", "Sort order"), SortOptions.Alphabetical
        )
        self.sort_selector.addItem(
            translate("AddonsInstaller", "Last updated", "Sort order"), SortOptions.LastUpdated
        )
        self.sort_selector.addItem(
            translate("AddonsInstaller", "Date created", "Sort order"), SortOptions.DateAdded
        )
        self.sort_selector.addItem(
            translate("AddonsInstaller", "GitHub stars", "Sort order"), SortOptions.Stars
        )
        if self.has_rankings:
            self.sort_selector.addItem(
                translate("AddonsInstaller", "Score", "Sort order"), SortOptions.Score
            )



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD Collada importer"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package importDAE
#  \ingroup ARCH
#  \brief DAE (Collada) file format importer and exporter
#
#  This module provides tools to import and export Collada (.dae) files.

from dataclasses import dataclass
import os
import string
from typing import Optional
from xml.sax.saxutils import escape as sax_escape

import numpy as np

import FreeCAD
import Arch
import Draft
import Mesh
import MeshPart

from draftutils import params

if FreeCAD.GuiUp:
    from draftutils.translate import translate
else:
    # \cond
    def translate(context, text):
        return text

    # \endcond


def xml_escape(text: str, entities: dict[str, str] = None) -> str:
    """Escape text for XML.

    This is a wrapper around xml.sax.saxutils.escape that replaces also
    `"` with `&quot;` by default.
    """
    if entities is None:
        entities = {'"': "&quot;"}
    return sax_escape(text, entities=entities)


def import_collada() -> bool:
    """Return True if the `collada` module is available.

    Also imports the module.

    """
    global collada
    try:
        import collada
    except ImportError:
        FreeCAD.Console.PrintError("pycollada not found, collada support is disabled.\n")
        return False
    return True


def triangulate(shape):
    """Triangulate the given shape."""
    mesher = params.get_param_arch("ColladaMesher")
    tessellation = params.get_param_arch("ColladaTessellation")
    grading = params.get_param_arch("ColladaGrading")
    segs_per_edge = params.get_param_arch("ColladaSegsPerEdge")
    segs_per_radius = params.get_param_arch("ColladaSegsPerRadius")
    second_order = params.get_param_arch("ColladaSecondOrder")
    optimize = params.get_param_arch("ColladaOptimize")
    allow_quads = params.get_param_arch("ColladaAllowQuads")
    if mesher == 0:
        return shape.tessellate(tessellation)
    elif mesher == 1:
        return MeshPart.meshFromShape(Shape=shape, MaxLength=tessellation).Topology
    else:
        return MeshPart.meshFromShape(
            Shape=shape,
            GrowthRate=grading,
            SegPerEdge=segs_per_edge,
            SegPerRadius=segs_per_radius,
            SecondOrder=second_order,
            Optimize=optimize,
            AllowQuad=allow_quads,
        ).Topology


def open(filename):
    """Called when FreeCAD wants to open a file."""
    if not import_collada():
        return
    docname = os.path.splitext(os.path.basename(filename))[0]
    doc = FreeCAD.newDocument(docname)
    doc.Label = docname
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def insert(filename, docname):
    """Called when FreeCAD wants to import a file."""
    if not import_collada():
        return
    try:
        doc = FreeCAD.getDocument(docname)
    except NameError:
        doc = FreeCAD.newDocument(docname)
    FreeCAD.ActiveDocument = doc
    read(filename)
    return doc


def read(filename):
    """Read a DAE file."""
    doc = FreeCAD.activeDocument()
    if not doc:
        return
    col = collada.Collada(filename, ignore=[collada.common.DaeUnsupportedError])
    # Read the unitmeter info from DAE file and compute unit to convert to mm.
    unit_meter = col.assetInfo.unitmeter or 1.0
    unit = unit_meter / 0.001
    bound_geom: collada.geometry.BoundGeometry
    # Implementation note: there's also `col.geometries` but when using them,
    # the materials are string and Gaël didn't find a way to get the material
    # node from this string.
    for bound_geom in col.scene.objects("geometry"):
        prim: collada.primitive.BoundPrimitive
        for prim in bound_geom.primitives():
            if isinstance(prim, collada.triangleset.BoundTriangleSet):
                _read_bound_triangle_set(doc, bound_geom, prim, unit)
            if isinstance(prim, collada.polylist.BoundPolylist):
                FreeCAD.Console.PrintWarning(
                    f"Warning: triangulating polylist primitive in {_name_from_bound_geom(bound_geom)}\n",
                )
                _read_bound_polylist(doc, bound_geom, prim, unit)
            # e.g. a BoundLineSet, which is not supported yet.

    # Print the errors that occurred during reading.
    if col.errors:
        FreeCAD.Console.PrintWarning("File was read but some errors occurred:\n")
    for e in col.errors:
        FreeCAD.Console.PrintWarning(str(e) + "\n")
    if FreeCAD.GuiUp:
        FreeCAD.Gui.SendMsgToActiveView("ViewFit")


def export(
    exports: list[FreeCAD.DocumentObject],
    filename: str,
    tessellation: int = 1,
    colors: Optional[dict[str, tuple]] = None,
):
    """Export FreeCAD contents to a DAE file.

    Parameters
    ----------
    - tessellation is used when breaking curved surfaces into triangles.
    - colors is an optional dictionary of {objName: shapeColorTuple} or
             {objName: diffuseColorList} elements to be used in non-GUI
             mode if you want to be able to export colors.

    """
    if not import_collada():
        return
    if colors is None:
        colors = {}
    scale = params.get_param_arch("ColladaScalingFactor")
    scale = scale * 0.001  # from millimeters (FreeCAD) to meters (Collada)
    default_color = Draft.get_rgba_tuple(params.get_param_view("DefaultShapeColor"))[:3]
    col_mesh = collada.Collada()
    col_mesh.assetInfo.upaxis = collada.asset.UP_AXIS.Z_UP
    # Authoring info.
    col_contributor = collada.asset.Contributor()
    try:
        author = FreeCAD.ActiveDocument.CreatedBy
    except UnicodeEncodeError:
        author = FreeCAD.ActiveDocument.CreatedBy.encode("utf8")
    author = xml_escape(author)
    col_contributor.author = author
    ver = FreeCAD.Version()
    appli = f"FreeCAD v{ver[0]}.{ver[1]} build {ver[2]}"
    col_contributor.authoring_tool = appli
    # Bug in collada from 0.4 to 0.9, contributors are not written to file.
    # Set it anyway for future versions.
    col_mesh.assetInfo.contributors.append(col_contributor)
    col_mesh.assetInfo.unitname = "meter"
    col_mesh.assetInfo.unitmeter = 1.0
    default_mat = None
    obj_ind = 0
    scene_nodes = []
    objects = Draft.get_group_contents(
        exports,
        walls=True,
        addgroups=True,
    )
    objects = Arch.pruneIncluded(objects, strict=True)
    for obj in objects:
        findex = np.array([])
        m: Optional[Mesh.Mesh] = None
        if obj.isDerivedFrom("Part::Feature"):
            FreeCAD.Console.PrintMessage(
                f'Exporting shape of object {obj.Name} ("{obj.Label}")' + "\n"
            )
            new_shape = obj.Shape.copy()
            new_shape.Placement = obj.getGlobalPlacement()
            m = Mesh.Mesh(triangulate(new_shape))
        elif obj.isDerivedFrom("Mesh::Feature"):
            FreeCAD.Console.PrintMessage(
                f'Exporting mesh of object {obj.Name} ("{obj.Label}")' + "\n"
            )
            m = obj.Mesh
        elif obj.isDerivedFrom("App::Part"):
            for child in obj.OutList:
                objects.append(child)
            continue
        else:
            continue
        if m:
            topology = m.Topology
            facets = m.Facets

            # Vertex indices.
            vindex = np.empty(len(topology[0]) * 3)
            for i in range(len(topology[0])):
                v = topology[0][i]
                vindex[list(range(i * 3, i * 3 + 3))] = (v.x * scale, v.y * scale, v.z * scale)

            # Normals.
            nindex = np.empty(len(facets) * 3)
            for i in range(len(facets)):
                n = facets[i].Normal
                nindex[list(range(i * 3, i * 3 + 3))] = (n.x, n.y, n.z)

            # Face indices.
            findex = np.empty(len(topology[1]) * 6, np.int64)
            for i in range(len(topology[1])):
                f = topology[1][i]
                findex[list(range(i * 6, i * 6 + 6))] = (f[0], i, f[1], i, f[2], i)

        vert_src = collada.source.FloatSource(f"cubeverts-array{obj_ind}", vindex, ("X", "Y", "Z"))
        normal_src = collada.source.FloatSource(
            f"cubenormals-array{obj_ind}", nindex, ("X", "Y", "Z")
        )
        geom = collada.geometry.Geometry(
            collada=col_mesh,
            id=f"geometry{obj_ind}",
            name=xml_escape(obj.Label),
            sourcebyid=[vert_src, normal_src],
        )
        input_list = collada.source.InputList()
        input_list.addInput(0, "VERTEX", f"#cubeverts-array{obj_ind}")
        input_list.addInput(1, "NORMAL", f"#cubenormals-array{obj_ind}")
        mat_node: Optional[collada.scene.MaterialNode] = None
        mat_ref = "materialref"
        if (
            hasattr(obj, "Material")
            and obj.Material
            and hasattr(obj.Material, "Material")
            and ("DiffuseColor" in obj.Material.Material)
        ):
            kd = tuple(
                [float(k) for k in obj.Material.Material["DiffuseColor"].strip("()").split(",")]
            )
            effect = collada.material.Effect(
                id=f"effect_{obj.Material.Name}",
                params=[],
                shadingtype="phong",
                diffuse=kd,
                specular=(1, 1, 1),
            )
            mat = collada.material.Material(
                id=f"mat_{obj.Material.Name}",
                name=obj.Material.Name,
                effect=effect,
            )
            col_mesh.effects.append(effect)
            col_mesh.materials.append(mat)
            mat_ref = f"ref_{obj.Material.Name}"
            mat_node = collada.scene.MaterialNode(
                symbol=mat_ref,
                target=mat,
                inputs=[],
            )
        if not mat_node:
            if obj.Name in colors:
                color = colors[obj.Name]
                if color:
                    if isinstance(color[0], tuple):
                        # This is a diffusecolor. For now, use the first color.
                        # TODO: Support per-face colors
                        color = color[0]
                    kd = color[:3]
                    effect = collada.material.Effect(
                        id=f"effect_{obj.Name}",
                        params=[],
                        shadingtype="phong",
                        diffuse=kd,
                        specular=(1, 1, 1),
                    )
                    mat = collada.material.Material(
                        id=f"mat_{obj.Name}",
                        name=obj.Name,
                        effect=effect,
                    )
                    col_mesh.effects.append(effect)
                    col_mesh.materials.append(mat)
                    mat_ref = f"ref_{obj.Name}"
                    mat_node = collada.scene.MaterialNode(
                        symbol=mat_ref,
                        target=mat,
                        inputs=[],
                    )
            elif FreeCAD.GuiUp:
                if hasattr(obj.ViewObject, "ShapeColor"):
                    kd = obj.ViewObject.ShapeColor[:3]
                    effect = collada.material.Effect(
                        id=f"effect_{obj.Name}",
                        params=[],
                        shadingtype="phong",
                        diffuse=kd,
                        specular=(1, 1, 1),
                    )
                    mat = collada.material.Material(
                        id=f"mat_{obj.Name}",
                        name=obj.Name,
                        effect=effect,
                    )
                    col_mesh.effects.append(effect)
                    col_mesh.materials.append(mat)
                    mat_ref = f"ref_{obj.Name}"
                    mat_node = collada.scene.MaterialNode(
                        symbol=mat_ref,
                        target=mat,
                        inputs=[],
                    )
        if not mat_node:
            if not default_mat:
                effect = collada.material.Effect(
                    id="effect_default",
                    params=[],
                    shadingtype="phong",
                    diffuse=default_color,
                    specular=(1, 1, 1),
                )
                default_mat = collada.material.Material(
                    id="mat_default",
                    name="default_material",
                    effect=effect,
                )
                col_mesh.effects.append(effect)
                col_mesh.materials.append(default_mat)
            mat_node = collada.scene.MaterialNode(
                symbol=mat_ref,
                target=default_mat,
                inputs=[],
            )
        triset = geom.createTriangleSet(indices=findex, inputlist=input_list, materialid=mat_ref)
        geom.primitives.append(triset)
        col_mesh.geometries.append(geom)
        geom_node = collada.scene.GeometryNode(geom, [mat_node])
        node = collada.scene.Node(id=f"node{obj_ind}", children=[geom_node])
        scene_nodes.append(node)
        obj_ind += 1
    scene = collada.scene.Scene("scene", scene_nodes)
    col_mesh.scenes.append(scene)
    col_mesh.scene = scene
    col_mesh.write(filename)
    FreeCAD.Console.PrintMessage(f'file "{filename}" successfully created.\n')


@dataclass
class _TriangleSet:
    """The result of the processing of BoundTriangleSet and BoundPolylist."""

    # List of triangles, each triangle is a list of 3 vertices,
    # each vertex is a list of 3 floats.
    triangles: list[list[float]]

    # The material associated to these triangles, or None.
    material: Optional["collada.material.Material"]


def _read_bound_triangle_set(
    doc: FreeCAD.Document,
    bound_geom: "collada.geometry.BoundGeometry",
    prim: "collada.primitive.BoundPrimitive",
    unit: float,
) -> list[FreeCAD.DocumentObject]:
    """
    Read a BoundTriangleSet primitive.

    Return the generated objects.

    Parameters
    ----------
    - bound_geom is the BoundGeometry containing the primitive.
    - prim is the BoundTriangleSet primitive to read.
    - unit is the factor to convert from the Collada unit to mm.
    """
    if not isinstance(prim, collada.triangleset.BoundTriangleSet):
        raise TypeError("Expected a BoundTriangleSet")
    triangle_sets = _get_triangle_sets_from_triangle(prim, unit)
    return _add_meshes_to_doc(doc, triangle_sets, _name_from_bound_geom(bound_geom))


def _read_bound_polylist(
    doc: FreeCAD.Document,
    bound_geom: "collada.geometry.BoundGeometry",
    prim: "collada.primitive.BoundPrimitive",
    unit: float,
) -> list[FreeCAD.DocumentObject]:
    """
    Read a BoundPolylist primitive.

    Return the generated objects.

    Parameters
    ----------
    - bound_geom is the BoundGeometry containing the primitive.
    - prim is the BoundPolylist primitive to read.
    - unit is the factor to convert from the Collada unit to mm.
    """
    if not isinstance(prim, collada.polylist.BoundPolylist):
        raise TypeError("Expected a BoundPolylist")
    triangle_sets = _get_triangle_sets_from_polylist(prim, unit)
    return _add_meshes_to_doc(
        doc, triangle_sets, _name_from_bound_geom(bound_geom) + " (triangulated)"
    )


def _get_triangle_sets_from_triangle(
    prim: "collada.triangleset.BoundTriangleSet",
    unit: float,
) -> list[_TriangleSet]:
    """Return triangles from the given BoundTriangleSet."""
    # Get the materials and associated triangles.
    material_map: dict[collada.scene.MaterialNode, list] = {}
    tri: collada.triangleset.Triangle
    for tri in prim:
        material_node: "collada.material.Material" = tri.material
        if material_node not in material_map:
            # Not yet in the dict, create a new entry.
            material_map[material_node] = []
        # tri.vertices is a numpy array.
        material_map[material_node].append((tri.vertices * unit).tolist())
    return [_TriangleSet(triangles=t, material=m) for m, t in material_map.items()]


def _get_triangle_sets_from_polylist(
    prim: "collada.triangleset.BoundPolylist",
    unit: float,
) -> list[_TriangleSet]:
    """Return triangles from the given BoundPolylist."""
    # Get the materials and associated triangles.
    material_map: dict[collada.scene.MaterialNode, list] = {}
    poly: collada.triangleset.Triangle
    for poly in prim:
        material_node: "collada.material.Material" = poly.material
        if material_node not in material_map:
            # Not yet in the dict, create a new entry.
            material_map[material_node] = []
        for tri in poly.triangles():
            # tri.vertices is a numpy array.
            material_map[material_node].append((tri.vertices * unit).tolist())
    return [_TriangleSet(triangles=t, material=m) for m, t in material_map.items()]


def _name_from_bound_geom(
    bound_geom: "collada.geometry.BoundGeometry",
) -> str:
    """Return `name` or `id` of the original geometry."""
    name = bound_geom.original.name
    if not name:
        name = bound_geom.original.id
    return name


def _add_meshes_to_doc(
    doc: FreeCAD.Document,
    triangle_sets: list[_TriangleSet],
    name: str,
) -> list[FreeCAD.DocumentObject]:
    """Create a mesh object for each _TriangleSet."""
    objects: list[FreeCAD.DocumentObject] = []
    for triangle_set in triangle_sets:
        mesh = Mesh.Mesh(triangle_set.triangles)
        obj = doc.addObject("Mesh::Feature", name)
        # Geometry name can be an ID, which often finishes with a number.
        # In order to prevent FreeCAD from incrementing the number (because
        # duplicate labels are not allowed by default) but adding
        # its "001", "002", .. suffices, add "_".
        suffix = "_" if name and (name[-1] in string.digits) else ""
        obj.Label = name + suffix
        # Label2 (i.e. Description) can be non-unique.
        obj.Label2 = name
        obj.Mesh = mesh
        objects.append(obj)
        if not triangle_set.material:
            continue
        if not FreeCAD.GuiUp:
            # Unfortunately, material can only be treated in GUI mode.
            continue
        fc_mat = FreeCAD.Material()
        # We do not import transparency because it is often set
        # wrongly (transparency mistaken for opacity).
        # TODO: Ask whether to import transparency.
        field_map = {
            "ambient": "AmbientColor",
            "diffuse": "DiffuseColor",
            "emission": "EmissiveColor",
            "specular": "SpecularColor",
            "shininess": "Shininess",
            # "transparency": "Transparency",
        }
        for col_field, fc_field in field_map.items():
            try:
                # Implementation note: using floats, so values must
                # be within [0, 1]. OK.
                setattr(fc_mat, fc_field, getattr(triangle_set.material.effect, col_field))
            except ValueError:
                # The collada value is not compatible with FreeCAD.
                pass
            except TypeError:
                # color is not a tuple but a texture.
                pass
        obj.ViewObject.ShapeAppearance = (fc_mat,)
    return objects
